/*
* MobiliTools: an implementation of the Object Management Group's
* Mobile Agent Facility specification.
* Copyright (C) 2003 France Telecom R&D
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
* MobiliTools $Name:  $
*
* Contact: bruno.dillenseger@francetelecom.com
*
* Authors: Bruno Dillenseger
*/


import org.objectweb.mobilitools.smi.api.Agency;
import org.objectweb.mobilitools.smi.api.*;
import java.applet.*;
import java.io.*;


/**
 * MobiliTools $Name:  $, $Id: SmiTest.java,v 1.1.1.1 2003/03/28 14:47:58 dillense Exp $
 * <P>
 * This example class shows MobileObject lifecycle through a console-based user interaction, as well as resource loading.
 * The easiest way to use this example is to use GUI-enabled agencies.
 * <BR>Note that this agent has no autonomous activity.
 */
public class SmiTest implements MobileObject
{
	/** display-friendly name, set once for all at creation */
	String my_name;
	/** true if the agent is currently active, false if it is suspended */
	boolean running;
	/** reference to current host agency, set after each move - must be transient because non-serializable */
	transient Agency my_agency;


	/**
	 * Default constructor is invoked when no argument is provided for agent creation.
	 */
	public SmiTest()
	{
		System.out.println("SmiTest default constructor");
	}


	/**
	 * (Generic) Example of specific constructor. Displays the argument provided
	 * for agent creation.
	 */
	public SmiTest(Object arg)
	{
		System.out.println("SmiTest constructor with argument " + arg);
	}


	//////////////////////////////////////////////
	// implementation of MobileObject interface //
	//////////////////////////////////////////////


	/**
	 * Displays a message and plays a sound on agent creation. Aks confirmation of creation.
	 * Answer 'y' or 'n' to respectively accept or deny agent creation. If creation is denied,
	 * the creation-caller gets an exception.
	 */
    public void afterBirth(AgentSystem agency, AgentInfo entry, Object argument)
		throws BadOperation
	{
		my_agency = (Agency)agency;
		my_name = new String(entry.getName().identity());
		System.out.println("agent " + (new Name(entry.getName().getStringRepresentation())).toString());
		running = true;
		playSound("12312.au");
		if (! confirm(my_name + ".afterBirth() is OK ([y]/n)?", true))
		{
			throw new BadOperation(BadOperation.REJECTED, "Agent creation is denied.");
		}
	}


	/**
	 * Asks confirmation before moving. Answer 'y' or 'n' in the console to respectively
	 * confirm or abort the move. In case you abort, an exception is raised, the agent
	 * is not serialized, and the move-caller gets an exception.
	 */
	public void beforeMove(Location location, String place)
		throws BadOperation
	{
		if (! confirm(
			my_name
			+ ".beforeMove("
			+ location.getRegion() + " "
			+ location.getAgency() + ", "
			+ place + ") is OK ([y]/n)?",
			true))
		{
			throw new BadOperation(BadOperation.REJECTED, "Agent move is denied.");
		}
	}


	/**
	 * Plays a sound and asks installation confirmation after deserialization in the target agency.
	 * Answer 'y' or 'n' to respectively accept or deny installation. Denying will drop the agent copy
	 * in the target agency and invoke <code>afterMoveFailed()</code> on the agent in source agency.
	 */
	public void afterMove(AgentSystem agency, Location location, String place)
		throws BadOperation
	{
		my_agency = (Agency)agency;
		playSound("12312.au");
		if (!confirm(
			my_name
			+ ".afterMove("
			+ location.getRegion() + " "
			+ location.getAgency() + ", "
			+ place + ") is OK ([y]/n)?",
			true))
		{
			throw new BadOperation(BadOperation.REJECTED, "Agent re-installation is denied.");
		}
	}


	/**
	 * Plays a sound and displays a message in the source agency after the agent has been serialized
	 * but its installation in the target agency has failed (probably rejected by the agent itself)
	 */
	public void afterMoveFailed(
		Location location,
		String place,
		int reason,
		String message)
	{
		playSound("12312.au");
		System.out.println(
			my_name
			+ ".afterMoveFailed("
			+ location.getRegion() + " "
			+ location.getAgency() + ", "
			+ place + ")");
		System.out.println("code " + String.valueOf(reason) + ": " + message);
	}


	/**
	 * Displays a message on host agency shutdown
	 */
	public void beforeShutdown()
	{
		System.out.println(my_name + ".beforeShutdown()");
	}


	/**
	 * Displays a message on termination
	 */
	public void beforeDeath()
	{
		System.out.println(my_name + ".beforeDeath()");
	}


	/**
	 * Displays an acknowledgement message on agent activity resumption if and only if it was suspended,
	 * otherwise the resume-caller gets an exception.
	 */
	public void beforeResume()
		throws BadOperation
	{
		if (running)
		{
			throw new BadOperation(BadOperation.RUNNING, "Agent activity was already running.");
		}
		running = true;
		System.out.println(my_name + ".resume()");
	}


	/**
	 * Displays an acknowledgement message on agent activity suspension if and only if it was active,
	 * otherwise the suspend-caller gets an exception.
	 */
	public void beforeSuspend()
		throws BadOperation
	{
		if (! running)
		{
			throw new BadOperation(BadOperation.RUNNING, "Agent activity was already suspended.");
		}
		running = false;
		System.out.println(my_name + ".suspend()");
	}


	/////////////////////////
	// own support methods //
	/////////////////////////


	/**
	 * Displays a message and reads a confirmation ('y' or 'n' whatever upper or lower case)
	 * in the attached console/terminal
	 * @param message confirmation question to be displayed
	 * @param defaultYes default answer (i.e. when neither 'y' nor 'n' is entered)
	 * @return true if the user answered 'y', false if the user answered 'n', defaultYes otherwise.
	 */
	boolean confirm(String message, boolean defaultYes)
	{
		try
		{
			System.out.print(message);
			String reply = (new BufferedReader(new InputStreamReader(System.in))).readLine();
			return reply.equalsIgnoreCase("y") || (defaultYes && ! reply.equalsIgnoreCase("n"));
		}
		catch (IOException e)
		{
			e.printStackTrace();
		}
		return defaultYes;
	}


	/**
	 * Loads a sound designated by the provided resource name using the agent's class classloader and plays it.
	 * @param resourcename the name of the resource (should be an audio file)
	 */
	void playSound(String resourcename)
	{
		try
		{
			Applet.newAudioClip(getClass().getClassLoader().getResource(resourcename)).play();
		}
		catch (Exception ex)
		{
			System.err.println("could not load and play audio resource " + resourcename + "\n" + ex);

		}
	}
}
