/*
* MobiliTools: an implementation of the Object Management Group's
* Mobile Agent Facility specification.
* Copyright (C) 2003 France Telecom R&D
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
* MobiliTools $Name:  $
*
* Contact: bruno.dillenseger@francetelecom.com
*
* Authors: Bruno Dillenseger
*/


import org.objectweb.mobilitools.smi.api.Agency;
import org.objectweb.mobilitools.smi.api.*;
import org.objectweb.mobilitools.smi.goodies.RegionManager;


/**
 * MobiliTools $Name:  $, $Id: ThreadAgent.java,v 1.1.1.1 2003/03/28 14:47:58 dillense Exp $
 * <P>
 * This example class shows a straightforward implementation of a thread-based agent activity.
 * The agent activity consists in continuously moving from agency to agency in its region.
 * The agent takes an optional single integer argument, interpreted as a delay in ms (no argument means no delay).
 * The easiest way to use this example is to use GUI-enabled agencies.
 * <BR>NB : THIS EXAMPLE REQUIRES THE MAFFINDER TO BE RUNNING BEFORE LAUNCHING AGENCIES AND AGENTS.
 * <P>
 * At creation, the agent gets the latest list of agencies.
 * While this list contains only one agency (it cannot be empty!), the agent loops on updating the list
 * and waiting for the specified delay. As soon as the list contains at least two agencies,
 * the agent begins to move, waiting for the specified delay between two consecutive moves.
 * <UL>
 * A move may fail for two reasons:
 * <LI>If the target agency no longer exists or it is currently closing, the agent gets the latest list of agencies.
 * <LI>If the agent's current agency is closing, the move is locally rejected and the agent terminates.
 * </UL>
 * If the update of agency list fails, the agent suspends itself.
 */
public class ThreadAgent implements MobileObject, Runnable
{
	/** agent's name */
	protected Name my_name;
	/** current host agency */
	protected transient Agency my_agency;
	/** latest list of agencies in current region */
	protected volatile Location[] itinerary;
	/** delay in ms between consecutive moves or agency list updates */
	protected int pause = 0;
	/** index in list of agencies for next move */
	protected int step;
	/** true if the agent is dead */
	protected boolean dead = false;
	/** true if the agent is suspended */
	protected boolean suspended = false;


	public ThreadAgent()
	{
	}


	/**
	 * Update the list of agencies and set the itinerary index to zero.
	 * Suspend the agent if the list cannot be updated.
	 */
	protected synchronized void resetItinerary()
	{
		try
		{
			itinerary = (new RegionManager(my_agency.getORB())).listAgencies(my_agency.getRegion(), null);
			step = 0;
		}
		catch (BadOperation ex)
		{
			System.out.println(my_name + " could not update itinerary - suspending activity");
			try
			{
				my_agency.suspendLocalAgent(this);
			}
			catch (BadOperation ex2)
			{
				System.err.println("Warning: could not suspend agent " + my_name + ":\n" + ex2);
			}
		}
	}


	/**
	 * Agent's thread creation.
	 */
	protected void runActivity()
	{
		(new Thread(this, my_name.toString())).start();
	}


	////////////////////////
	// interface Runnable //
	////////////////////////


	/**
	 * Agent activity: wait for a while, then move to next agency in itinerary
	 * or update itinerary if it contains a single agency.
	*/
	public void run()
	{
		try
		{
			Thread.sleep(pause);
		}
		catch (InterruptedException e)
		{
			System.err.println("Warning: interrupted sleep for " + my_name);
		}
		synchronized (this)
		{
			if (! (dead || suspended))
			{
				if (itinerary.length == 1)
				{
					resetItinerary();
					runActivity();
				}
				else
				{
					if (step == itinerary.length)
					{
						step = 0;
					}
					try
					{
						my_agency.moveLocalAgent(this, itinerary[step++], "a_place");
					}
					catch (BadOperation e)
					{
						System.err.println(my_name + ": " + e.getMessage());
						resetItinerary();
						runActivity();
					}
				}
			}
		}
	}


	////////////////////////////
	// interface MobileObject //
	////////////////////////////


	/**
	 * Initialize the agent and create a thread to run the agent's autonomous activity
	 * @param arg if a non-empty String[], should hold a delay in ms as a single value;
	 * this delay is applied between two consecutive moves or itinerary updates.
	 * Otherwise, this delay is zero.
	 */
	public void afterBirth(AgentSystem agency, AgentInfo entry, Object arg)
		throws BadOperation
	{
		my_agency = (Agency)agency;
		my_name = entry.getName();
		if (arg instanceof String[])
		{
			String[] argStr = (String[])arg;
			if (argStr.length != 0)
			{
				try
				{
					pause = Integer.parseInt(argStr[0]);
				}
				catch (NumberFormatException e)
				{
					throw new BadOperation(BadOperation.REJECTED, "argument should be an String-formated integer", e);
				}
			}
		}
		resetItinerary();
		runActivity();
	}


	/**
	 * Make the agent's thread stop as soon as possible.
	 */
	public synchronized void beforeSuspend()
	{
		System.out.println(my_name + " is suspended");
		suspended = true;
	}


	/**
	 * Update the itinerary in current agency's region, and create a thread
	 * to run the agent's autonomous activity.
	 */
	public synchronized void beforeResume()
		throws BadOperation
	{
		suspended = false;
		resetItinerary();
		runActivity();
	}


	/**
	 * Nothing special to do before moving.
	 */
	public void beforeMove(Location dummy1, String dummy2)
	{
	}


	/**
	 * Create a thread after move to go on agent's autonomous activity in new host agency.
	 */
	public void afterMove(AgentSystem agency, Location dummy1, String dummy2)
	{
		my_agency = (Agency)agency;
		runActivity();
	}


	/**
	 * Nothing special to do if the agent transport/setup phase failed during a move.
	 */
	public void afterMoveFailed(
		Location dummy1,
		String dummy2,
		int dummy3,
		String dummy4)
	{
	}


	/**
	 * Make the agent's thread stop as soon as possible.
	 */
	public synchronized void beforeDeath()
	{
		dead = true;
		System.out.println(my_name + " is dead.");
	}


	/**
	 * Nothing special to do before agency shutdown (just face upcoming death...)
	 */
	public void beforeShutdown()
	{
	}
}
