/*
* MobiliTools: an implementation of the Object Management Group's
* Mobile Agent Facility specification.
* Copyright (C) 2003 France Telecom R&D
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
* MobiliTools $Name:  $
*
* Contact: mobilitools-smi@lists.debian-sf.objectweb.org
*
* Authors: Bruno Dillenseger
*/


package org.objectweb.mobilitools.smi;


import java.util.Observable;
import java.util.TreeMap;
import java.util.SortedMap;
import java.util.Iterator;
import java.util.Vector;
import java.util.Properties;
import java.util.Enumeration;
import org.objectweb.mobilitools.smi.api.Constants;
import org.objectweb.mobilitools.smi.idl.*;
import org.objectweb.mobilitools.util.corba.NameService;
import org.objectweb.mobilitools.util.corba.NameServiceException;
import org.omg.CfMAF.*;
import org.omg.CORBA.ORB;
import org.omg.CORBA.Any;


/**
 * MobiliTools $Name:  $, $Id: Finder.java,v 1.1.1.1 2003/03/28 14:48:05 dillense Exp $
 * <P>
 * Each instance launches a new implementation of the MAFFinder interface for a given region.
 * The class may also be directly run to launch a single MAFFinder.
 * In SMI, MAFFinder implementations rely on the CORBA naming service to publish their object
 * reference.
 * Only one MAFFinder can be launched for a given region.
 * @see Constants
 * @see MAFFinder_impl
 * @see org.objectweb.mobilitools.smi.goodies.FinderGUI
*/
public class Finder extends Observable
{
	static boolean standalone = false;


	/**
		Launches a MAFFinder for the given region, and registers its object
		reference in the default CORBA naming service.
		@param args the last argument must specify the region name
		@see #Finder(String, ORB)
	*/
	static public void main(String args[])
	{
		ORB orb = ORB.init(args, null);
		if (args.length == 0)
		{
			System.err.println("argument required: region_name");
			System.exit(-1);
		}
		else
		{
			standalone = true;
			new Finder(args[args.length-1], orb);
			System.out.println("MAFFinder for region " + args[0] + " is ready.");
			orb.run();
		}
	}


	// string'fied name in the CORBA naming service
	String my_ns_name;

	// name of the region managed by current Finder
	String my_region;

	// object implementing the MAFFinder interface
	_MAFFinderComplementImplBase my_finderTie;

	// frontend to CORBA naming service
	NameService my_ns;

	// reference to associated ORB object
	ORB my_orb;

	// ordered list of Finder entries (agents and agent systems)
	SortedMap entries;


	/**
		Launches a MAFFinder for the given region, and registers its object
		reference in the default CORBA naming service.
		@param region the name of the region to manage.
		@see Constants
	*/
	public Finder(String region)
	{
		this(region, ORB.init(new String[0], System.getProperties()));
	}


	/**
		Launches a MAFFinder for the given region, and registers its object
		reference in the default CORBA naming service.
		@param region the name of the region to manage.
		@param orb the ORB object to use.
		@see Constants
		@see org.omg.CORBA.ORB#init()
	*/
	public Finder(String region, ORB orb)
	{
		entries = new TreeMap();
		my_region = region;
		my_orb = orb;
		my_finderTie = new MAFFinder_impl(this);
		orb.connect(my_finderTie);
		my_ns_name =
			System.getProperty(Constants.regionPrefixProp, Constants.regionPrefixDefault) +
			"/" +
			region +
			"/" +
			System.getProperty(Constants.finderNameProp, Constants.finderNameDefault);
		try
		{
			my_ns = new NameService(orb);
			my_ns.rebind(my_ns_name, my_finderTie);
		}
		catch (NameServiceException e)
		{
			e.printStackTrace();
			System.err.println("Exception with naming service: unable to bind name " + my_ns_name);
			System.err.println(e.toString());
		}
	}


	/**
		@return the name of the region managed by current Finder.
	*/
	public String getRegion()
	{
		return my_region;
	}


	/**
		Shuts the finder.
		@return 0 if the MAFFinder object reference could be canceled from the CORBA naming service,
		-1 otherwise.
	*/
	public int exit()
	{
		int exitcode = -1;
		try
		{
			my_ns.unbind(my_ns_name);
			my_orb.disconnect(my_finderTie);
			exitcode = 0;
		}
		catch (NameServiceException exception)
		{
			System.err.println(exception.toString());
		}
		if (standalone)
		{
			System.exit(exitcode);
		}
		return(exitcode);
	}


	///////////////////////////////////////////////////////////
	// actual implementation of MAFFinder-related operations //
	///////////////////////////////////////////////////////////


	public void register(FinderEntry entry)
		throws NameInvalid
	{
		org.objectweb.mobilitools.smi.api.Name key = entry.getName();
		if (key.getmafname().identity.length == 0)
		{
			throw new NameInvalid();
		}
		synchronized(entries)
		{
			if (entries.containsKey(key))
			{
				if (entry.isAgent())
				{
					entries.remove(key);
				}
				else
				{
					throw new NameInvalid();
				}
			}
			entries.put(key, entry);
			if (countObservers() > 0)
			{
				setChanged();
				try
				{
					notifyObservers(entries.values().toArray(new FinderEntry[0]));
				}
				catch (Throwable e)
				{
					e.printStackTrace();
				}
			}
		}
	}


	public Name[] lookup_names(Object profile)
	{
		Name[] result;
		Object names = lookup(null, null, profile, true);
		if (names instanceof Name[])
		{
			result = (Name[])names;
		}
		else
		{
			result = new Name[0];
		}
		return result;
	}


	public String[] lookup_locations(org.objectweb.mobilitools.smi.api.Name name, String location, Object profile)
		throws EntryNotFound
	{
		String[] result;
		Object locations = lookup(name, location, profile, false);
		if (locations instanceof String[])
		{
			result = (String[])locations;
			if (result.length == 0)
			{
				throw new EntryNotFound();
			}
		}
		else
		{
			throw new EntryNotFound();
		}
		return result;
	}


	private Object lookup(org.objectweb.mobilitools.smi.api.Name name, String location, Object profile, boolean returnNames)
	{
		Object result = null;
		if (name == null || name.getmafname().identity.length == 0)
		{
			synchronized (entries)
			{
				Iterator all = entries.values().iterator();
				Vector select = new Vector();
				while (all.hasNext())
				{
					FinderEntry element = (FinderEntry)all.next();
					if (Misc.matchProfile(profile, element.getProfile()) &&
						(location == null || location.length() == 0 || location.equals(element.getLocation())))
					{
						if (returnNames)
						{
							select.add(element.getName().getmafname());
						}
						else
						{
							select.add(element.getLocation());
						}
					}
				}
				if (returnNames)
				{
					result = new Name[select.size()];
					result = (Name[])(select.toArray((Name[])result));
				}
				else
				{
					result = new String[select.size()];
					result = (String[])select.toArray((String[])result);
				}
			}
		}
		else
		{
			FinderEntry element = (FinderEntry)entries.get(name);
			if (element != null)
			{
				result = new String[] { element.getLocation() };
			}
		}
		return result;
	}


	public void unregister(org.objectweb.mobilitools.smi.api.Name name)
		throws EntryNotFound
	{
		if (entries.remove(name) == null)
		{
			throw new EntryNotFound();
		}
		else if (countObservers() > 0)
		{
			setChanged();
			notifyObservers(entries.values().toArray(new FinderEntry[0]));
		}
	}


	public AgentProfile getAgentProfile(org.objectweb.mobilitools.smi.api.Name name)
		throws AgentNotFound
	{
		AgentProfile result = null;
		FinderEntry entry = (FinderEntry)entries.get(name);
		if (entry != null && entry.getProfile() instanceof AgentProfile)
		{
			result = (AgentProfile)entry.getProfile();
		}
		if (result == null)
		{
			throw new AgentNotFound();
		}
		else
		{
			return result;
		}
	}


	public void printEntries()
	{
		Iterator values = entries.values().iterator();
		while (values.hasNext())
		{
			FinderEntry entry = (FinderEntry)values.next();
			System.out.println(entry.toString() + ", location " + entry.getLocation());
		}
	}


	public ORB getORB()
	{
		return my_orb;
	}
}
