/*
* MobiliTools: an implementation of the Object Management Group's
* Mobile Agent Facility specification.
* Copyright (C) 2003 France Telecom R&D
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
* MobiliTools $Name:  $
*
* Contact: mobilitools-smi@lists.debian-sf.objectweb.org
*
* Authors: Bruno Dillenseger
*/


package org.objectweb.mobilitools.smi;


import org.omg.CORBA.*;
import org.omg.CfMAF.*;
import java.lang.reflect.*;
import java.lang.Object;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;

import org.objectweb.mobilitools.smi.api.*;
import org.objectweb.mobilitools.smi.idl.*;
import org.objectweb.mobilitools.smi.lib.ClassCache;


/**
 * MobiliTools $Name:  $, $Id: MAFAgentSystem_impl.java,v 1.1.1.1 2003/03/28 14:48:05 dillense Exp $
 * <P>
 * Implementation of MAF's MAFAgentSystem interface (always bound to an Agency instance).
 * @see Agency
*/
public class MAFAgentSystem_impl extends _MAFAgentSystemComplementImplBase
	implements MAFAgentSystemOperations, MAFAgentSystemComplementOperations
{
	Agency my_agency;
	boolean trace;


	/**
		Creates a new implementation of the MAFAgentSystem MAF interface.
		Every invocation is (more or less directly) delegated to a SMI agency.
		@param agency the bound Agency instance.
	*/
	public MAFAgentSystem_impl(Agency agency)
	{
		this(agency, false);
	}


	/**
		Creates Creates a new implementation of the MAFAgentSystem MAF interface.
		Every invocation is (more or less directly) delegated to a SMI agency.
		@param agency the bound Agency instance.
		@param traceOn must be true to switch call trace on
	*/
	public MAFAgentSystem_impl(Agency agency, boolean traceOn)
	{
		my_agency = agency;
		trace = traceOn;
	}


	/////////////////////////////////////////////////////////
	// implementation of MAF's MAFAgentSystem operations //
	/////////////////////////////////////////////////////////


	/**
		Creates an agent.
	*/
	public org.omg.CfMAF.Name create_agent(
		org.omg.CfMAF.Name agent_name,
		AgentProfile agent_profile,
		byte[] agent,
		String place_name,
		byte[] arguments,
		ClassName[] class_names,
		String code_base,
		MAFAgentSystem class_provider)
		throws ClassUnknown, ArgumentInvalid, DeserializationFailed, MAFExtendedException
	{
		if (trace)
		{
			System.out.println("MAFAgentSystem::create_agent()");
		}
		if (agent_profile.agent_system_type != my_agency.getAgentSystemInfo().agent_system_type)
		{
			throw new MAFExtendedException();
		}
		try
		{
  			ObjectInputStream deserializer = new ObjectInputStream(
				new ByteArrayInputStream(arguments));
			java.lang.Object argObj = deserializer.readObject();
			return my_agency.createLocalAgent(
				class_names[0].name,
				new org.objectweb.mobilitools.smi.api.Name(agent_name),
				agent_profile,
				place_name,
				code_base,
				agent,
				argObj,
				class_provider).getName().getmafname();
		}
		catch (IOException e)
		{
			throw new DeserializationFailed();
		}
		catch (ClassNotFoundException e)
		{
			throw new ClassUnknown();
		}
		catch (BadOperation ex)
		{
			System.err.println(ex.getMessage());
			switch (ex.getReason())
			{
			case BadOperation.CLASSFAULT:
				throw new ClassUnknown();
			case BadOperation.SERIALIZATION:
				throw new DeserializationFailed();
			case BadOperation.REJECTED:
				throw new ArgumentInvalid();
			default:
				throw new MAFExtendedException();
			}
		}
	}


	/**
		Returns code for requested classes.
	*/
	public byte[][] fetch_class(
		ClassName[] class_name_list,
		String code_base,
		AgentProfile agent_profile)
		throws ClassUnknown, MAFExtendedException
	{
		if (trace)
		{
			System.out.println("MAFAgentSystem::fetch_class()");
		}
		byte[][] code = new byte[class_name_list.length][];
		ClassCache cache = ClassCache.getCache(code_base);
		synchronized(cache)
		{
			for (int i=0 ; i<code.length ; ++i)
			{
				code[i] = (byte[])cache.get(class_name_list[i].name);
				if (code[i] == null)
				{
					System.err.println("unknown class " + class_name_list[i].name + " for codebase " + code_base);
					System.out.println(cache);
					throw new ClassUnknown();
				}
			}
		}
		return code;
	}


	/**
		Not implemented.
	*/
	public String find_nearby_agent_system_of_profile(
		AgentProfile agent_profile)
		throws EntryNotFound, NO_IMPLEMENT
	{
		if (trace)
		{
			System.out.println("MAFAgentSystem::find_nearby_agent_system_of_profile() - NOT IMPLEMENTED !");
		}
		throw new NO_IMPLEMENT("Not implemented");
	}


	/**
		@return the activity status of an agent, i.e. running or suspended.
		Note that the CfMAFTerminated status is never returned (the exception
		AgentNotFound would be thrown is case of a terminated agent).
		@exception AgentNotFound if the agent name does not exist.
	*/
	public AgentStatus get_agent_status(
		org.omg.CfMAF.Name agent_name)
		throws AgentNotFound
	{
		if (trace)
		{
			System.out.println("MAFAgentSystem::get_agent_status()");
		}
		try
		{
			if (my_agency.localAgentIsRunning(new org.objectweb.mobilitools.smi.api.Name(agent_name)))
			{
				return AgentStatus.CfMAFRunning;
			}
			else
			{
				return AgentStatus.CfMAFSuspended;
			}
		}
		catch (BadOperation ex)
		{
			throw new AgentNotFound();
		}
	}


	/**
		@return a information record about the agency.
	*/
	public AgentSystemInfo get_agent_system_info()
	{
		if (trace)
		{
			System.out.println("MAFAgentSystem::get_agent_system_info()");
		}
		return my_agency.getAgentSystemInfo();
	}


	/**
		Not implemented.
	*/
	public AuthInfo get_authinfo(
		org.omg.CfMAF.Name agent_name)
		throws AgentNotFound, NO_IMPLEMENT
	{
		if (trace)
		{
			System.out.println("MAFAgentSystem::get_authinfo() - NOT IMPLEMENTED !");
		}
		throw new NO_IMPLEMENT("Not implemented");
	}


	/**
		@return the CORBA object reference to the MAFFinder.
		@exception FinderNotFound current agency is running without MAFFinder support.
	*/
	public MAFFinder get_MAFFinder()
		throws FinderNotFound
	{
		if (trace)
		{
			System.out.println("MAFAgentSystem::get_MAFFinder()");
		}
		MAFFinder result = my_agency.getMAFFinder();
		if (result != null)
		{
			return result;
		}
		else
		{
			throw new FinderNotFound();
		}
	}


	/**
		@return the list of all agents residing in the associated agency.
	*/
	public org.omg.CfMAF.Name[] list_all_agents()
	{
		if (trace)
		{
			System.out.println("MAFAgentSystem::list_all_agents()");
		}
		return my_agency.listAgents((AgentProfile)null);
	}


	/**
		@return an array of agent names with given authority.
	*/
	public org.omg.CfMAF.Name[] list_all_agents_of_authority(
		byte[] authority)
	{
		if (trace)
		{
			System.out.println("MAFAgentSystem::list_all_agents_of_authority()");
		}
		return my_agency.listAgents(authority);
	}


	/**
		Not implemented.
	*/
	public String[] list_all_places()
		throws NO_IMPLEMENT
	{
		if (trace)
		{
			System.out.println("MAFAgentSystem::list_all_places() - NOT IMPLEMENTED !");
		}
		throw new NO_IMPLEMENT("Not implemented");
	}


	/**
		Receives a migrating agent.
	*/
	public void receive_agent(
		org.omg.CfMAF.Name agent_name,
		AgentProfile agent_profile,
		byte[] agent,
		String place_name,
		ClassName[] class_names,
		String code_base,
		MAFAgentSystem sender)
		throws ClassUnknown, DeserializationFailed, MAFExtendedException
	{
		if (trace)
		{
			System.out.println("MAFAgentSystem::receive_agent()");
		}
		try
		{
			my_agency.receiveAgent(
				class_names[0].name,
				agent,
				agent_name,
				agent_profile,
				place_name,
				code_base,
				sender);
		}
		catch (ClassUnknown e)
		{
			throw e;
		}
		catch (DeserializationFailed e)
		{
			throw e;
		}
		catch (MAFExtendedException e)
		{
			throw e;
		}
		catch (Throwable e)
		{
			throw new BAD_OPERATION(e.toString());
		}
	}


	/**
		Resumes the activity of an agent.
	*/
	public void resume_agent(
		org.omg.CfMAF.Name agent_name)
		throws AgentNotFound, ResumeFailed, AgentIsRunning
	{
		if (trace)
		{
			System.out.println("MAFAgentSystem::resume_agent()");
		}
		try
		{
			my_agency.resumeLocalAgent(new org.objectweb.mobilitools.smi.api.Name(agent_name));
		}
		catch (BadOperation ex)
		{
			switch (ex.getReason())
			{
			case BadOperation.RUNNING:
				throw new AgentIsRunning();
			case BadOperation.UNKNOWNAGENT:
				throw new AgentNotFound();
			default:
				throw new ResumeFailed();
			}
		}
	}


	/**
		Suspends the activity of an agent.
	*/
	public void suspend_agent(
		org.omg.CfMAF.Name agent_name)
		throws AgentNotFound, SuspendFailed, AgentIsSuspended
	{
		if (trace)
		{
			System.out.println("MAFAgentSystem::suspend_agent()");
		}
		try
		{
			my_agency.suspendLocalAgent(new org.objectweb.mobilitools.smi.api.Name(agent_name));
		}
		catch (BadOperation ex)
		{
			switch (ex.getReason())
			{
			case BadOperation.SUSPENDED:
				throw new AgentIsSuspended();
			case BadOperation.UNKNOWNAGENT:
				throw new AgentNotFound();
			default:
				throw new SuspendFailed();
			}
		}
	}


	/**
		Terminates an agent.
	*/
	public void terminate_agent(
		org.omg.CfMAF.Name agent_name)
		throws AgentNotFound, TerminateFailed
	{
		if (trace)
		{
			System.out.println("MAFAgentSystem::terminate_agent()");
		}
		try
		{
			my_agency.terminateLocalAgent(new org.objectweb.mobilitools.smi.api.Name(agent_name));
		}
		catch (BadOperation ex)
		{
			System.err.println(ex.getMessage());
			if (ex.getReason() == BadOperation.UNKNOWNAGENT)
			{
				throw new AgentNotFound();
			}
		}
	}


	/**
		Terminates the agency.
	*/
	public void terminate_agent_system()
		throws TerminateFailed
	{
		if (trace)
		{
			System.out.println("MAFAgentSystem::terminate_agent_system()");
		}
		my_agency.terminate();
	}


	/////////////////////////////////////////////////////////////////
	// implementation of SMI's MAFAgentSystemComplement operations //
	/////////////////////////////////////////////////////////////////


	public org.omg.CfMAF.Name[] list_agents(AgentProfile profile)
	{
		if (trace)
		{
			System.out.println("MAFAgentSystemComplement::list_agents()");
		}
		return my_agency.listAgents(profile);
	}


	public void move_agent(org.omg.CfMAF.Name name, String destination, String place)
		throws
		AgentNotFound,
		ClassUnknown,
		DeserializationFailed,
		AgentSystemNotFound,
		ArgumentInvalid,
		AgentRefusedToMove,
		MAFExtendedException
	{
		if (trace)
		{
			System.out.println("MAFAgentSystemComplement::move_agent()");
		}
		try
		{
			my_agency.moveLocalAgent(
				new org.objectweb.mobilitools.smi.api.Name(name),
				new Location(destination),
				place);
		}
		catch (BadOperation e)
		{
			switch (e.getReason())
			{
			case BadOperation.REJECTED:
				throw new AgentRefusedToMove();
			case BadOperation.DESTINATION:
				throw new AgentSystemNotFound();
			case BadOperation.CLASSFAULT:
				throw new ClassUnknown();
			case BadOperation.UNKNOWNAGENT:
				throw new AgentNotFound();
			case BadOperation.SERIALIZATION:
				throw new DeserializationFailed();
			case BadOperation.INFRASTRUCTURE:
				throw new MAFExtendedException();
			}
		}
	}
}
