/*
* MobiliTools: an implementation of the Object Management Group's
* Mobile Agent Facility specification.
* Copyright (C) 2003 France Telecom R&D
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
* MobiliTools $Name:  $
*
* Contact: mobilitools-smi@lists.debian-sf.objectweb.org
*
* Authors: Bruno Dillenseger
*/


package org.objectweb.mobilitools.smi.api;


import org.omg.CORBA.ORB;
import org.omg.CORBA.Any;
import org.omg.CORBA.SystemException;
import org.omg.CORBA.UnknownUserException;
import org.omg.CORBA.CompletionStatus;
import org.omg.CosNaming.NamingContextPackage.*;
import org.omg.CfMAF.*;
import org.objectweb.mobilitools.util.corba.NameService;
import org.objectweb.mobilitools.util.corba.NameServiceException;
import org.objectweb.mobilitools.util.thread.ShutdownLock;
import org.objectweb.mobilitools.smi.idl.*;
import org.objectweb.mobilitools.smi.*;
import java.util.*;
import java.io.*;
import java.lang.reflect.*;


/**
    MobiliTools $Name:  $, $Id: Agency.java,v 1.1.1.1 2003/03/28 14:48:05 dillense Exp $
    <P>
	Instances of this class are execution environments for agents, supporting
	lifecycle/mobility management, using and providing MAFAgentSystem implementation.
	An agency may be created by directly instantiating this class, by running this class
	from command line or by running the AgencyGUI class.
	<P>
	An agency may be customized with an <code>AgencyPersonality</code> implementation. A <I>personality</I>
	may be declared when creating a new agency (using the appropriate constructor), or via
	inheritance. A personality's main role is to encapsulate agents by intercepting all lifecycle callbacks defined by
	interface MobileObject.
	@see org.objectweb.mobilitools.smi.goodies.AgencyGUI
	@see AgencyPersonality
	@see MobileObject
*/
public class Agency
	extends Observable
	implements AgentSystem, AgencyPersonality
{
	/**
		if using an agency personality, set this attribute to true if you wish
		the agency to perform System.exit(0) when terminating the agency.
	*/
	static protected boolean standalone = false;
	static public final String NOTIFY_TERMINATE = "terminate";

	/**
		Launches a SMI agency.
		@param args command-line arguments should specify:
		<UL>
			<LI>args[0] the agency unique name in the given region,
			<LI>args[1] the region name.
		</UL>
		@see #Agency(String, String, ORB)
	*/
	public static void main(String[] args)
	{
		ORB orb = ORB.init(args, null);
		String region = null;
		standalone = true;
		if (args.length >= 2)
		{
			try
			{
				new Agency(args[0], args[1], orb);
			}
			catch (BadOperation e)
			{
				System.err.println("Name " + args[0] + " already exists in region " + args[1] + ".");
				System.exit(-2);
			}
		}
		else
		{
			System.err.println("arguments: agency_name region_name");
			System.exit(-1);
		}
		System.out.println("Agency " + args[0] + " available in region " + args[1] + ".");
		orb.run();
	}


	// region's name
	protected String my_region;

	// agency's name
	protected String my_name;

	// object reference to the MAFAgentSystem object
	protected _MAFAgentSystemComplementImplBase tie;

	// naming service front-end
	protected NameService my_ns;

	// name of the MAFAgentSystem object in the name service
	protected String my_ns_name;

	// location of the agency (CosNaming URI format)
	protected Location my_location;

	// reference to the ORB object
	protected ORB my_orb;

	// hashed lists of hosted agents
	protected Hashtable agentsByRef, agentsByName;

	// CORBA object reference to MAFFinder
	protected MAFFinder my_finder;

	// reference to an agency personality (if any)
	protected AgencyPersonality my_personality;

	// reference to the hosting agent system
	protected AgentSystem my_shell;

	// information record about the agent system (see MAF)
	protected AgentSystemInfo my_info;

	// lock to enforce a clean agency shutdown
	protected ShutdownLock shutdownLock;

	/**
		Create a new agency for creating and receiving agents.
		The authority for this agency is set from the "user.name" system
		property. The agency's MAFAgentSystem interface is bound in the
		CORBA naming service accordingly to the region name and the
		agency name.
		@param name name of the agency, unique in the given region.
		@param region name of the region the agency belongs to.
		@throws BadOperation if the name of the agency is not accepted by the MAFFinder (name already used?).
	*/
	public Agency(String name, String region)
		throws BadOperation
	{
		this(name, region, ORB.init(new String[0], System.getProperties()));
	}


	/**
		Create a new agency for creating and receiving agents, and
		customize it with the provided agency personality and shell.
		The agency's MAFAgentSystem interface is bound in the
		CORBA naming service accordingly to the region name and the
		agency name.
		@param region name of the region the agency belongs to.
		@param personality object intercepting agent lifecycle callbacks.
		@param shell reference that will be given to the agents
		created in or moving to this agency.
		@param info data record describing the customized agency (see MAF).
		This info should hold the agency's MAF name, whose identity field is taken as the
		agency's unique name in SMI's sense.
		@throws BadOperation if the name of the agency is not accepted by the MAFFinder (name already used).
		@see Constants#makeAgentSystemInfo(byte[], byte[])
	*/
	public Agency(
		String region,
		AgencyPersonality personality,
		AgentSystem shell,
		AgentSystemInfo info)
		throws BadOperation
	{
		this(region,
			ORB.init(new String[0], System.getProperties()),
			personality,
			shell,
			info);
	}


	/**
		Create a new agency for creating and receiving agents.
		The authority for this agency is set from the "user.name" system
		property. The agency's MAFAgentSystem interface is bound in the
		CORBA naming service accordingly to the region name and the
		agency name.
		@param name name of the agency, unique in the given region.
		@param region name of the region the agency belongs to.
		@param orb the ORB object to use.
		@throws BadOperation if the name of the agency is not accepted by the MAFFinder (name already used?).
		@see org.omg.CORBA.ORB#init()
	*/
	public Agency(String name, String region, ORB orb)
		throws BadOperation
	{
		this(
			region,
			orb,
			null,
			null,
			Constants.makeAgentSystemInfo(
				((String)System.getProperty("user.name", "-unspecified-")).getBytes(),
				name.getBytes()));
	}


	/**
		Create a new agency for creating and receiving agents.
		The authority for this agency is set from the "user.name" system
		property. The agency's MAFAgentSystem interface is bound in the
		CORBA naming service accordingly to the region name and the
		agency name.
		@param region name of the region the agency belongs to.
		@param orb the ORB object to use.
		@param personality the agency personality that will handle the agents. A null value means
		no personality (i.e. current agency instance).
		@param shell agency reference that will be given to the hosted agents after creation or move.
		A null value means no specific shell, in which case the host agency will be used.
		@param info the AgentSystemInfo information record about the agent system personality.
		This info should hold the agency's MAF name, whose identity field is taken as the
		agency's unique name in SMI's sense.
		@throws BadOperation if the name of the agency is not accepted by the MAFFinder (already used?).
		@see org.omg.CORBA.ORB#init()
		@see Constants#makeAgentSystemInfo(byte[], byte[])
	*/
	public Agency(
		String region,
		ORB orb,
		AgencyPersonality personality,
		AgentSystem shell,
		AgentSystemInfo info)
		throws BadOperation
	{
		shutdownLock = new ShutdownLock();
		agentsByRef = new Hashtable();
		agentsByName = new Hashtable();
		my_info = info;
		my_name = new String(info.agent_system_name.identity);
		my_region = region;
		my_location = new Location(region, my_name);
		my_ns_name = my_location.getCosNamingName();
		my_orb = orb;
		if (personality == null)
		{
			my_personality = this;
		}
		else
		{
			my_personality = personality;
		}
		if (shell == null)
		{
			my_shell = this;
		}
		else
		{
			my_shell = shell;
		}

		// create a MAFAgentSystem implementation, connect it to the ORB,

		tie = new MAFAgentSystem_impl(this, false);
		my_orb.connect(tie);
		String finder_name =
			System.getProperty(Constants.regionPrefixProp, Constants.regionPrefixDefault) +
			"/" +
			region +
			"/" +
			System.getProperty(Constants.finderNameProp, Constants.finderNameDefault);
		// get CORBA naming service
		try
		{
			my_ns = new NameService(my_orb);
		}
		catch (NameServiceException e)
		{
			System.err.println("Fatal error - could not get CORBA name service: " + e);
			throw new BadOperation(BadOperation.INFRASTRUCTURE, "Cannot create agency without COS name service", e);
		}
		// get the MAF Finder of agency's region (if any)
		try
		{
			my_finder = MAFFinderHelper.narrow(my_ns.resolve(finder_name));
			my_finder.register_agent_system(
				my_info.agent_system_name,
				my_location.toString(),
				my_info);
		}
		catch (NameServiceException e)
		{
			System.err.println("Unable to reach the MAFFinder in " + finder_name + " - no region finder support: " + e);
			my_finder = null;
		}
		catch (SystemException e)
		{
			System.err.println("Unable to reach the MAFFinder " + finder_name + " - no region finder support: " + e);
			my_finder = null;
		}
		catch (NameInvalid e)
		{
			throw new BadOperation(BadOperation.INVALIDNAME, "Exception in agency creation: name " + new Name(my_info.agent_system_name) + " has been rejected by the MAFFinder", e);
		}
		// bind the agency name in CORBA naming service
		try
		{
			if (my_finder == null)
			{
				my_ns.bind(my_ns_name, tie);
			}
			else // force binding since the MAFFinder has accepted the name
			{
				my_ns.rebind(my_ns_name, tie);
			}
		}
		catch (NameServiceException e)
		{
			System.err.println("Exception with CORBA naming service - unable to bind name " + my_ns_name + ":" + e);
			if (my_finder != null)
			{
				try
				{
					my_finder.unregister_agent_system(my_info.agent_system_name);
				}
				catch (Exception ex)
				{
					System.err.println("Warning - could not unregister agency " + new Name(my_info.agent_system_name) + " from the MAFFinder while agency creation is aborted: " + e);
				}
			}
			throw new BadOperation(BadOperation.INVALIDNAME, "Could not bind " + my_ns_name + " in CORBA naming service.", e);
		}
	}


	/**
		@return a data record describing current agent system (see MAF).
	*/
	public AgentSystemInfo getAgentSystemInfo()
	{
		return my_info;
	}


	/**
		@return a new AgentProfile instance for SMI (or personality) agents,
		with the given properties.
		@param properties the agent properties to be inserted in the profile.
	*/
	public AgentProfile makeAgentProfile(Properties properties)
	{
		return new AgentProfile(
			Constants.LANGUAGE_ID,
			my_info.agent_system_type,
			my_info.agent_system_description,
			my_info.major_version,
			my_info.minor_version,
			Constants.SERIALIZATION,
			Misc.property2any(properties, my_orb));
	}


	/**
		Receives a moving agent,
		invokes callback <code>afterMove()</code> on the newly installed agent.
		@param class_name class name of the received agent.
		@param agent_state the state of the received agent.
		@param agent_name the MAF unique name of the received agent.
		@param agent_profile the profile of the received agent (see MAF specifications).
		@param place the name of the place where to settle the new agent.
		@param code_base codebase to load classes for this agent.
		@param class_provider the MAFAgentSystem implementation from where to pick the classes
		of the received agent.
		@throws ClassUnknown at least one of the needed classes could not be found
		@throws DeserializationFailed the received agent is not of SMI agent system type,
		or an exception was thrown during the agent deserialization
		@throws MAFExtendedException an exception was thrown by the agent afterMove() method
	*/
	public void receiveAgent(
		String class_name,
		byte[] agent_state,
		org.omg.CfMAF.Name agent_name,
		AgentProfile agent_profile,
		String place,
		String code_base,
		MAFAgentSystem class_provider)
		throws ClassUnknown, DeserializationFailed, MAFExtendedException
	{
		try
		{
			shutdownLock.in();
		}
		catch (ShutdownLock.ShutdownException e)
		{
			throw Constants.agencyIsClosed;
		}
		try
		{
			// check agent system type Id
			if (agent_profile.agent_system_type != my_info.agent_system_type)
			{
				System.err.println("Warning: rejected attempt to receive an object of foreign agent system type (" + String.valueOf(agent_name.agent_system_type) + ").");
				throw new DeserializationFailed();
			}
			// deserialization of the incoming agent
			Boolean isRunning;
			MobileObject agent = null;
			try
			{
				SMIObjectInputStream deserializer = new SMIObjectInputStream(
					new ByteArrayInputStream(agent_state),
					code_base,
					agent_profile,
					class_provider);
				isRunning = (Boolean) deserializer.readObject();
				agent = (MobileObject) deserializer.readObject();
			}
			catch (IOException e)
			{
				e.printStackTrace();
				throw new DeserializationFailed();
			}
			catch (ClassNotFoundException e)
			{
				e.printStackTrace();
				throw new ClassUnknown();
			}
			// registration of the new location of the agent in the MAFFinder (if available)
			if (my_finder != null)
			{
				try
				{
					my_finder.register_agent(agent_name, my_location.toString(), agent_profile);
				}
				catch (SystemException e)
				{
					e.printStackTrace();
					throw e;
				}
				catch (NameInvalid e)
				{
					throw new MAFExtendedException();
				}
			}
			// add the new agent to agent tables, and invoke callback on MobileObject
			Name agent_name_wrapper = new Name(agent_name);
			AgentInfo entry = new AgentEntry(
				agent,
				new Name(agent_name),
				place,
				isRunning.booleanValue(),
				code_base,
				Misc.any2property(agent_profile.properties));
			synchronized(agent)
			{
				registerAgent(agent, agent_name_wrapper, entry);
				try
				{
					my_personality.afterMove(agent, my_shell, my_location, place);
				}
				catch (Exception e)
				{
					if (!(e instanceof BadOperation))
					{
						System.err.println("Warning: unexpected exception thrown by agent " + agent_name_wrapper + " in afterMove()");
						e.printStackTrace();
					}
					unregisterAgent(agent, agent_name_wrapper);
					throw new MAFExtendedException();
				}
			}
			setChanged();
			notifyObservers();
		}
        finally
		{
			shutdownLock.out();
		}
	}


	/**
		@return CORBA object reference to MAFFinder implementation this agency is bound to.
		Returns null if the agency is running without MAFFinder support.
	*/
	public MAFFinder getMAFFinder()
	{
		return my_finder;
	}


	/**
		@return the name of the region the agency belongs to.
	*/
	public String getRegion()
	{
		return my_region;
	}


	/**
		@return a wrapper for the MAF name of the agency.
	*/
	public Name getName()
	{
		return new Name(my_info.agent_system_name);
	}


	/**
		@return the full name of the agency, as it appears in the CORBA naming service.
		String'fied COSnaming names have the following format:
		<code>/naming context1!kind1/.../object!kind</code>
	*/
	public String getNsName()
	{
		return my_ns_name;
	}


	/**
		@return the location of the agency.
	*/
	public Location getLocation()
	{
		return my_location;
	}


	/**
		@return a record of data about an agent hosted by the agency, or null
		if the designated agent is not known by the agency.
		@param agent the MAF unique name of the agent.
	*/
	public AgentInfo getAgentInfo(Name agent)
	{
		AgentInfo result;
		result = (AgentInfo)agentsByName.get(agent);
		return result;
	}


	/**
		@return a record of data about an agent hosted by the agency, or null
		if the designated agent is not known by the agency.
		@param agent the Java object reference to the agent.
	*/
	public AgentInfo getAgentInfo(MobileObject agent)
	{
		synchronized(agent)
		{
			AgentInfo result;
			result = (AgentInfo)agentsByRef.get(agent);
			return result;
		}
	}


	/**
		Creates an agent in current agency.
		The agent's default constructor is called unless a constructor taking
		a single Object as argument is defined and "argument" is both non-null and
		not an empty array.
		@param agent_class the class name of the new agent.
		@param name the identity part of the MAF name of the new agent
		(must be unique in the set of names of the agency's authority; the authority
		part of the MAF agent name will be the same as the agency's).
		@param place the name of the place where to settle the new agent.
		@param codebase codebase to load classes for this agent.
		@param properties agent's properties.
		@param argument parameter for agent constructor and initialization.
		@throws BadOperation the agent creation failed
		@see MobileObject#afterBirth(AgentSystem, AgentInfo, Object)
	*/
	public AgentInfo createLocalAgent(
		String agent_class,
		Name name,
		String place,
		String codebase,
		Properties properties,
		Object argument)
		throws BadOperation
	{
		return createLocalAgent(
			agent_class,
			name,
			makeAgentProfile(properties),
			place,
			codebase,
			null,
			argument,
			null);
	}


	/**
		Creates an agent in current agency.
		The agent's default constructor is called unless a constructor taking
		a single Object as argument is defined and "argument" is both non-null and
		not an empty array.
		@param classname the class name of the new agent.
		@param name the MAF unique name of the new agent.
		@param profile agent MAF profile.
		@param place the name of the place where to settle the new agent.
		@param codebase codebase to load classes for this agent.
		@param agent_data ignored if null or empty array ; otherwise, should contain a serialized
		Java object which will be used as agent constructor and initialization parameter.
		@param argument parameter for agent constructor and initialization (ignored if agent_data
		is non-null and non-empty).
		@param class_provider the MAFAgentSystem implementation from where to pick the classes;
		(if null, the codebase will be interpreted locally.
		@throws BadOperation the agent creation failed
		@see MobileObject#afterBirth(AgentSystem, AgentInfo, Object)
	*/
	public AgentInfo createLocalAgent(
		String classname,
		Name name,
		AgentProfile profile,
		String place,
		String codebase,
		byte[] agent_data,
		Object argument,
		MAFAgentSystem class_provider)
		throws BadOperation
	{
		Class agentClass = null;
		ClassLoader loader;
		try
		{
			loader = Misc.getClassLoader(
				Thread.currentThread().getContextClassLoader(),
				codebase,
				profile,
				class_provider);
			agentClass = loader.loadClass(classname);
		}
		catch (ClassNotFoundException e)
		{
			throw new BadOperation(BadOperation.CLASSFAULT, "Exception in create operation for agent " + name + ": undefined class " + classname + ".", e);
		}
		catch (Throwable e)
		{
			throw new BadOperation(BadOperation.INFRASTRUCTURE, "Exception in create operation for agent " + name + ": " + e);
		}
		if (agentClass == null)
		{
			throw new BadOperation(BadOperation.CLASSFAULT, "Exception in create operation for agent " + name + ": undefined class " + classname + ".");
		}
		if (agent_data != null && agent_data.length != 0)
		{
			try
			{
				SMIObjectInputStream deserializer = new SMIObjectInputStream(
					new ByteArrayInputStream(agent_data),
					loader);
				argument = deserializer.readObject();
			}
			catch (IOException e)
			{
				throw new BadOperation(BadOperation.SERIALIZATION, "Exception in create operation for agent " + name + ": could not deserialize argument object.", e);
			}
			catch (ClassNotFoundException e)
			{
				throw new BadOperation(BadOperation.CLASSFAULT, "Exception in create operation for agent " + name + ": undefined class " + classname + ".", e);
			}
		}
		return createLocalAgent(
			agentClass,
			name,
			profile,
			place,
			codebase,
			argument);
	}


	/**
		Creates an agent in current agency, provided that its class has already
		been loaded (or can be loaded by the system classloader).
		The agent's default constructor is called unless a constructor taking
		a single Object as argument is defined and "argument" is both non-null and
		not an empty array.
		@param agent_class the class of the new agent.
		@param agent_name the MAF unique name of the new agent.
		@param agent_profile the profile of the agent (see MAF specifications).
		@param place the name of the place where to settle the new agent.
		@param codebase codebase to load classes for this agent.
		@param argument parameter for agent initialization and constructor.
		@throws BadOperation agent creation has failed
		@see MobileObject#afterBirth(AgentSystem, AgentInfo, Object)
		@see AgencyPersonality
	*/
	AgentInfo createLocalAgent(
		Class agent_class,
		Name agent_name,
		AgentProfile agent_profile,
		String place,
		String codebase,
		Object argument)
		throws BadOperation
	{
		try
		{
			shutdownLock.in();
		}
		catch (ShutdownLock. ShutdownException e)
		{
			throw Constants.agencyIsClosed;
		}
		try
		{
			MobileObject agent = null;
			AgentInfo entry;
			// check the agent system type of the agent
			if (agent_name.type() != my_info.agent_system_type)
			{
				throw new BadOperation(BadOperation.AGENTTYPE, "Exception in create operation for agent " + agent_name + ": unsupported agent system type " + String.valueOf(agent_name.type()) + ".");
			}

			// create a new instance and register the new agent in the MAFFinder (if available)
			try
			{
				if (argument != null
					&& ( ! (argument instanceof Object[])
						 || (argument instanceof Object[] && ((Object[])argument).length != 0)))
				{
					try
					{
						Constructor constructor = agent_class.getConstructor(new Class[]{ Object.class });
						agent = (MobileObject) constructor.newInstance(new Object[]{ argument });
					}
					catch (NoSuchMethodException ex)
					{
						agent = null;
					}
				}
				if (agent == null)
				{
					agent = (MobileObject) agent_class.newInstance();
				}
				entry = new AgentEntry(
					agent,
					agent_name,
					place,
					true,
					codebase,
					Misc.any2property(agent_profile.properties));
				if (my_finder != null)
				{
					my_finder.register_agent(
						agent_name.getmafname(),
						my_location.toString(),
						agent_profile);
				}
			}
			catch (InvocationTargetException e)
			{
				throw new BadOperation(BadOperation.REJECTED, "Exception in create operation for agent " + agent_name + ": exception in constructor.", e);
			}
			catch (InstantiationException e)
			{
				throw new BadOperation(BadOperation.INSTANTIATION, "Exception in create operation for agent " + agent_name, e);
			}
			catch (IllegalAccessException e)
			{
				throw new BadOperation(BadOperation.INSTANTIATION, "Exception in create operation for agent " + agent_name, e);
			}
			catch (ExceptionInInitializerError e)
			{
				throw new BadOperation(BadOperation.INSTANTIATION, "Exception in create operation for agent " + agent_name + ": " + e);
			}
			catch (SecurityException e)
			{
				throw new BadOperation(BadOperation.INSTANTIATION, "Exception in create operation for agent " + agent_name, e);
			}
			catch (SystemException e)
			{
				throw new BadOperation(BadOperation.INFRASTRUCTURE, "Exception in create operation for agent " + agent_name, e);
			}
			catch (NameInvalid e)
			{
				throw new BadOperation(BadOperation.OTHER, "Exception in create operation for agent " + agent_name + " (unspecified).", e);
			}
			// add agent to agent tables and invoke callback on MobileObject
			synchronized(agent)
			{
				registerAgent(agent, agent_name, entry);
				try
				{
					my_personality.afterBirth(agent, my_shell, entry, argument);
				}
				catch (Exception e)
				{
					unregisterAgent(agent, agent_name);
					if (my_finder != null)
					{
						try
						{
							my_finder.unregister_agent(agent_name.getmafname());
						}
						catch (SystemException ex)
						{
							ex.printStackTrace();
						}
						catch (EntryNotFound ex)
						{
							ex.printStackTrace();
						}
					}
					if (e instanceof BadOperation)
					{
						throw (BadOperation) e;
					}
					else
					{
						throw new BadOperation(BadOperation.REJECTED, "unexpected exception in afterBirth: " + e, e);
					}
				}
			}
			setChanged();
			notifyObservers();
			return entry;
		}
        finally
		{
			shutdownLock.out();
		}
	}


	/**
		Invokes the <code>beforeMove()</code> callback on the agent, and then tries
		to move it if no exception has been thrown by <code>beforeMove()</code>.
		If the move fails, the <code>afterMoveFailed()</code> callback is invoked
		on the agent, and an exception is thrown.
		@param agent name wrapper for agent's MAF name.
		@param agency location of the target agency.
		@param place the name of the target place in the given agency.
		@throws BadOperation on aborted move.
		@see MobileObject
		@see AgencyPersonality
	*/
	public void moveLocalAgent(Name agent, Location agency, String place)
		throws BadOperation
	{
		AgentInfo entry = (AgentInfo)agentsByName.get(agent);
		if (entry == null)
		{
			throw new BadOperation(
				BadOperation.UNKNOWNAGENT,
				"Exception in move operation: unknown agent " + agent + ".");
		}
		else
		{
			moveLocalAgent(entry.getAgent(), agency, place);
		}
	}


	/**
		Invokes the <code>beforeMove()</code> callback on the agent, and then tries
		to move it if no exception has been thrown by <code>beforeMove()</code>.
		If the move fails, the <code>afterMoveFailed()</code> callback is invoked
		on the agent, and an exception is thrown.
		@param agent Java object reference to the agent.
		@param agency location of the target agency.
		@param place the name of the target place in the given agency.
		@throws BadOperation on aborted move.
		@see MobileObject
		@see AgencyPersonality
	*/
	public void moveLocalAgent(MobileObject agent, Location agency, String place)
		throws BadOperation
	{
        try
		{
			shutdownLock.in();
		}
		catch (ShutdownLock.ShutdownException e)
		{
			throw Constants.agencyIsClosed;
		}
		try
		{
			synchronized(agent)
			{
				// check if the agent is known (in the agent tables)
				AgentInfo entry = (AgentInfo) agentsByRef.get(agent);
				if (entry == null)
				{
					throw new BadOperation(
						BadOperation.UNKNOWNAGENT,
						"Exception in move operation: unknown agent.");
				}
				// find destination MAFAgentSystem implementation in the naming service
				MAFAgentSystem dest_ref;
				String to_ns_name = agency.getCosNamingName();
				try
				{
					dest_ref = MAFAgentSystemHelper.narrow(my_ns.resolve(to_ns_name));
				}
				catch (NameServiceException e)
				{
					if (e.exception instanceof NotFound)
					{
						throw new BadOperation(
							BadOperation.DESTINATION,
							"Exception in move operation for agent " + entry.getName() + ": unknown destination agency " + agency + ".");
					}
					else
					{
						e.printStackTrace();
						throw new BadOperation(
							BadOperation.INFRASTRUCTURE,
							"Exception in move operation for agent " + entry.getName() + ": something is going wrong with CORBA naming service.");
					}
				}

				// try to move the agent:
				// - callbacks on MobileObject interface,
				// - remove the agent from agent tables,
				// - agent serialization,
				// - agent move (invocation of destination MAFAgentSystem implementation)
				// - according to overall migration success or failure, callbacks on
				// MobileObject interface.
				AgentProfile profile = makeAgentProfile(entry.getProperties());
				BadOperation exception = null;
				Boolean isRunning;
				try
				{
					my_personality.beforeMove(agent, agency, place);
				}
				catch (BadOperation ex)
				{
					throw ex;
				}
				catch (Exception ex)
				{
					throw new BadOperation(BadOperation.REJECTED, "Unexpected exception in beforeMove() for agent " + entry.getName(), ex);
				}
				isRunning = new Boolean(localAgentIsRunning(agent));
				try
				{
					unregisterAgent(agent, entry.getName());
					ClassName[] classes = {
						new ClassName(agent.getClass().toString().substring("class ".length()), new byte[0])
					};
					ByteArrayOutputStream agent_state = new ByteArrayOutputStream();
					ObjectOutputStream objOut = new ObjectOutputStream(agent_state);
					objOut.writeObject(isRunning);
					objOut.writeObject(agent);
					dest_ref.receive_agent(
						entry.getName().getmafname(),
						profile,
						agent_state.toByteArray(),
						place,
						classes,
						entry.getCodebase(),
						tie);
					setChanged();
					notifyObservers();
				}
				catch (IOException e)
				{
					exception = new BadOperation(BadOperation.SERIALIZATION, "Exception in move operation for agent " + entry.getName() + " (serialization)", e);
				}
				catch (SystemException e)
				{
					exception = new BadOperation(BadOperation.INFRASTRUCTURE, "Exception in move operation for agent " + entry.getName() + " (CORBA)", e);
				}
				catch (ClassUnknown e)
				{
					exception = new BadOperation(BadOperation.CLASSFAULT, "Exception in move operation for agent " + entry.getName() + " (missing class)", e);
				}
				catch (DeserializationFailed e)
				{
					exception = new BadOperation(BadOperation.CLASSFAULT, "Exception in move operation for agent " + entry.getName() + " (deserialization)", e);
				}
				catch (MAFExtendedException e)
				{
					exception = new BadOperation(BadOperation.REJECTED, "Exception in move operation for agent " + entry.getName() + " (agent refused to settle in destination agency)", e);
					try
					{
						my_finder.register_agent(entry.getName().getmafname(), my_location.toString(), profile);
					}
					catch (SystemException ex)
					{
						e.printStackTrace();
					}
					catch (NameInvalid ex)
					{
						e.printStackTrace();
					}
				}
				catch (ArgumentInvalid e)
				{
					exception = new BadOperation(BadOperation.OTHER, "Exception in move operation for agent " + entry.getName() + " (unspecified problem)", e);
				}
				if (exception != null)
				{
					registerAgent(agent, entry.getName(), entry);
					try
					{
						my_personality.afterMoveFailed(agent, agency, place, exception.getReason(), exception.getMessage());
					}
					catch (Exception ex)
					{
						System.err.println("Warning: ignoring unexpected exception thrown by agent " + entry.getName() + " in afterMoveFailed() call-back:\n" + ex);
						ex.printStackTrace();
					}
					throw exception;
				}
			}
		}
        finally
		{
			shutdownLock.out();
		}
	}


	/**
		Calls <code>beforeShutdown()</code> on the hosted agents,
		terminates agents that are still hosted, then calls <code>agencyShutdown()</code>
		on the agency, and finally shuts the agency down.
		@see MobileObject
		@see AgencyPersonality
	*/
	public void terminate()
	{
        try
		{
			shutdownLock.shutdown();
		}
		catch (ShutdownLock.ShutdownException e)
		{
			throw Constants.agencyIsClosed;
		}
		// unregister the agency from the MAFFinder (if available)
		if (my_finder != null)
		{
			try
			{
				my_finder.unregister_agent_system(my_info.agent_system_name);
			}
			catch (SystemException e)
			{
				System.err.println("Warning: unable to unregister agency from MAFFinder:\n" + e.toString());
			}
			catch (EntryNotFound e)
			{
				System.err.println("Warning: unable to unregister agency from MAFFinder:\n" + e.toString());
			}
		}
		synchronized (agentsByRef)
		{
			Enumeration agents;
			// warns all the hosted agents
			agents = agentsByRef.elements();
			while (agents.hasMoreElements())
			{
				AgentInfo entry = (AgentInfo)agents.nextElement();
				try
				{
					my_personality.beforeShutdown(entry.getAgent());
				}
				catch(Throwable ex)
				{
					System.err.println("Warning: ignored exception thrown in beforeShutdown() by agent " + entry.getName() + ".\n" + ex);
				}
			}
			// unbind the agency from the COS naming service and disconnect the agency
			try
			{
				my_ns.unbind(my_ns_name);
			}
			catch (NameServiceException e)
			{
				System.err.println("Warning: unable to unbind name " + my_ns_name + " from naming service:\n" + e.toString());
			}
			my_orb.disconnect(tie);
			// terminates all the hosted agents
			agents = agentsByRef.elements();
			while (agents.hasMoreElements())
			{
				try
				{
					terminateLocalAgent(((AgentInfo)agents.nextElement()).getName());
				}
				catch (BadOperation ex)
				{
					System.err.println("Exception in terminate operation for agency " + new Name(my_info.agent_system_name) + ":\n" + ex);
				}
			}
			my_personality.agencyShutdown();
		}
		// notify observers
		setChanged();
		notifyObservers(NOTIFY_TERMINATE);
		// trigger a delayed System.exit if the agency is "standalone"
		if (standalone)
		{
			new ExitDelay(Integer.parseInt(System.getProperty(Constants.agencyExitDelayProp, Constants.agencyExitDelayDefault)));
		}
	}


	/**
		Invokes the <code>beforeDeath()</code> callback on an agent, and then terminates it.
		@param agent MAF unique name of the agent.
		@throws BadOperation the target agent is not known by the agency, or
		its termination threw an exception.
		@see MobileObject
		@see AgencyPersonality
	*/
	public void terminateLocalAgent(Name agent)
		throws BadOperation
	{
		AgentInfo entry = (AgentInfo)agentsByName.get(agent);
		if (entry == null)
		{
			throw new BadOperation(BadOperation.UNKNOWNAGENT, "Exception in terminate operation: unknown agent " + agent + ".");
		}
		terminateLocalAgent(entry.getAgent());
	}


	/**
		Invokes the beforeDeath() callback on an agent, and then terminates it.
		@param agent the Java object reference to the agent.
		@throws BadOperation the target agent is not known by the agency, or
		its termination threw an exception.
		@see MobileObject
		@see AgencyPersonality
	*/
	public void terminateLocalAgent(MobileObject agent)
		throws BadOperation
	{
        try
		{
			shutdownLock.in();
		}
		catch (ShutdownLock.ShutdownException e)
		{
			throw Constants.agencyIsClosed;
		}
		try
		{
			BadOperation exception = null;
			synchronized(agent)
			{
				// check if the agent is known (in agent tables)
				AgentInfo entry = (AgentInfo)agentsByRef.get(agent);
				if (entry == null)
				{
					throw new BadOperation(BadOperation.UNKNOWNAGENT, "Exception in terminate operation: unknown agent.");
				}
				// callbacks on interface MobileObject
				// agent tables update
				// MAFFinder update (if available)
				try
				{
					my_personality.beforeDeath(agent);
				}
				catch (Throwable e)
				{
					System.err.println("Warning in terminateLocalAgent(" + entry.getName() + "): ignored exception thrown by beforeDeath().\n" + e.toString());
					exception = new BadOperation(
						BadOperation.REJECTED,
						"Unexpected exception in beforeDeath() call-back in agent " + entry.getName(),
						e instanceof Exception ? (Exception) e : null);
				}
				unregisterAgent(agent, entry.getName());
				try
				{
					if (my_finder != null)
					{
						my_finder.unregister_agent(entry.getName().getmafname());
					}
				}
				catch (Exception e)
				{
					System.err.println("Warning in terminateLocalAgent(" + entry.getName() + "): ignored exception dealing with MAFFinder.\n" + e);
					if (exception == null)
					{
						exception = new BadOperation(
							BadOperation.INFRASTRUCTURE,
							"Exception in terminate operation on agent " + entry.getName(),
							e);
					}
				}
			}
			setChanged();
			notifyObservers();
			if (exception != null)
			{
				throw exception;
			}
		}
		finally
		{
			shutdownLock.out();
		}
	}


	/**
		@return an array of names of hosted agents matching a given profile.
		The returned array is zero-sized if no agent matches the given profile (or if there is no agent at all).
		@param profile the agent profile to be matched; if null, all the agents are selected.
	*/
	public Name[] listLocalAgents(AgentProfile profile)
	{
		synchronized (agentsByRef)
		{
			Name[] result;
			if (profile == null)
			{
				result = new Name[agentsByName.size()];
				Enumeration keys = agentsByName.keys();
				for (int i=0 ; keys.hasMoreElements() ; ++i)
				{
					result[i] = (Name)keys.nextElement();
				}
			}
			else
			{
				Enumeration keys = agentsByName.keys();
				Vector select = new Vector();
				while (keys.hasMoreElements())
				{
					Name name = (Name)keys.nextElement();
					if (Misc.matchProfile(
						profile,
						makeAgentProfile(
							((AgentInfo)agentsByName.get(name)).getProperties())))
					{
						select.add(name);
					}
				}
				result = new Name[select.size()];
				result = (Name[])(select.toArray(result));
			}
			return result;
		}
	}


	/**
		@return an array of MAF names of hosted agents matching a given profile.
		The returned array is zero-sized if no agent matches the given profile (or if there is no agent at all).
		@param profile the agent profile to be matched; if null, all the agents are
		selected.
	*/
	public org.omg.CfMAF.Name[] listAgents(AgentProfile profile)
	{
		synchronized (agentsByRef)
		{
			org.omg.CfMAF.Name[] result;
			if (profile == null)
			{
				result = new org.omg.CfMAF.Name[agentsByName.size()];
				Enumeration keys = agentsByName.keys();
				for (int i=0 ; keys.hasMoreElements() ; ++i)
				{
					result[i] = ((Name)keys.nextElement()).getmafname();
				}
			}
			else
			{
				Enumeration keys = agentsByName.keys();
				Vector select = new Vector();
				while (keys.hasMoreElements())
				{
					Name name = (Name)keys.nextElement();
					if (Misc.matchProfile(
						profile,
						makeAgentProfile(
							((AgentInfo)agentsByName.get(name)).getProperties())))
					{
						select.add(name.getmafname());
					}
				}
				result = new org.omg.CfMAF.Name[select.size()];
				result = (org.omg.CfMAF.Name[])(select.toArray(result));
			}
			return result;
		}
	}


	/**
		@return an array of MAF names of local agents belonging to the given authority.
		The returned array is zero-sized if no agent matches the given profile (or if there is no agent at all).
		@param authority the agent's authority.
	*/
	public org.omg.CfMAF.Name[] listAgents(byte[] authority)
	{
		synchronized (agentsByRef)
		{
			org.omg.CfMAF.Name[] result;
			Enumeration keys = agentsByName.keys();
			Vector select = new Vector();
			while (keys.hasMoreElements())
			{
				Name name = (Name)keys.nextElement();
				if ((new String(authority)).equals(new String(name.authority())))
				{
					select.add(name.getmafname());
				}
			}
			result = new org.omg.CfMAF.Name[select.size()];
			result = (org.omg.CfMAF.Name[])(select.toArray(result));
			return result;
		}
	}


	/**
		Invokes the <code>beforeResume()</code> callback on an agent in order to resume its activity.
		Unless this callback throws an exception, the agent activity is then considered as resumed.
		@param agent MAF unique agent name.
		@throws BadOperation the agent could not be found and resumed, or it was already running.
		@see MobileObject
		@see AgencyPersonality
	*/
	public void resumeLocalAgent(Name agent)
		throws BadOperation
	{
		AgentInfo entry = (AgentInfo)agentsByName.get(agent);
		if (entry != null)
		{
			resumeLocalAgent(entry.getAgent());
		}
		else
		{
			throw new BadOperation(BadOperation.UNKNOWNAGENT, "Exception in resumeLocalAgent(): unknown agent " + agent + ".");
		}
	}


	/**
		Invokes the <code>beforeResume()</code> callback on an agent in order to resume its activity.
		Unless this callback throws an exception, the agent activity is then considered as resumed.
		@param agent the agent's Java object reference.
		@throws BadOperation the agent could not be found and resumed, or it was already running.
		@see MobileObject
		@see AgencyPersonality
	*/
	public void resumeLocalAgent(MobileObject agent)
		throws BadOperation
	{
		try
		{
			shutdownLock.in();
		}
		catch (ShutdownLock.ShutdownException e)
		{
			throw Constants.agencyIsClosed;
		}
		try
		{
			synchronized(agent)
			{
				AgentInfo entry = (AgentInfo)agentsByRef.get(agent);
				if (entry != null)
				{
					if (entry.isRunning())
					{
						throw new BadOperation(BadOperation.RUNNING, "Exception in resume operation: agent " + entry.getName() + " was already running.");
					}
					else
					{
						try
						{
							my_personality.beforeResume(agent);
						}
						catch (BadOperation ex)
						{
							throw ex;
						}
						catch (Exception ex)
						{
							throw new BadOperation(BadOperation.REJECTED, "Unexpected exception in beforeResume() for agent " + entry.getName(), ex);
						}
						entry.toggleRunning();
					}
				}
				else
				{
					throw new BadOperation(BadOperation.UNKNOWNAGENT, "Exception in resumeLocalAgent: unknown agent " + entry.getName() + ".");
				}
			}
		}
		finally
		{
			shutdownLock.out();
		}
	}


	/**
		Invokes the <code>beforeSuspend()</code> callback on an agent in order to suspend its activity.
		Unless this callback throws an exception, the agent activity is then considered as suspended.
		@param agent MAF unique agent name.
		@throws BadOperation the agent could not be found and suspended, or was already suspended.
		@see MobileObject
		@see AgencyPersonality
	*/
	public void suspendLocalAgent(Name agent)
		throws BadOperation
	{
		AgentInfo entry = (AgentInfo)agentsByName.get(agent);
		if (entry != null)
		{
			suspendLocalAgent(entry.getAgent());
		}
		else
		{
			throw new BadOperation(BadOperation.UNKNOWNAGENT, "Exception in suspendLocalAgent operation: unknown agent " + agent + ".");
		}
	}


	/**
		Invokes the <code>beforeSuspend()</code> callback on an agent in order to suspend its activity.
		Unless this callback throws an exception, the agent activity is then considered as suspended.
		@param agent the agent's Java object reference.
		@throws BadOperation the agent could not be found and suspended, or was already suspended.
		@see MobileObject
		@see AgencyPersonality
	*/
	public void suspendLocalAgent(MobileObject agent)
		throws BadOperation
	{
        try
		{
			shutdownLock.in();
		}
		catch (ShutdownLock.ShutdownException e)
		{
			throw Constants.agencyIsClosed;
		}
		try
		{
			synchronized(agent)
			{
				AgentInfo entry = (AgentInfo)agentsByRef.get(agent);
				if (entry != null)
				{
					if (! entry.isRunning())
					{
						throw new BadOperation(BadOperation.SUSPENDED, "Exception in suspend operation: agent " + entry.getName() + " was already suspended.");
					}
					else
					{
						try
						{
							my_personality.beforeSuspend(agent);
						}
						catch (BadOperation ex)
						{
							throw ex;
						}
						catch (Exception ex)
						{
							throw new BadOperation(BadOperation.REJECTED, "Unexpected exception in beforeSuspend() for agent " + entry.getName(), ex);
						}
						entry.toggleRunning();
					}
				}
				else
				{
					throw new BadOperation(BadOperation.UNKNOWNAGENT, "Exception in suspendLocalAgent(): unknown agent.");
				}
			}
		}
		finally
		{
			shutdownLock.out();
		}
	}


	/**
		Sets, changes or unsets an agent property.
		@param agent the agent MAF name wrapper.
		@param property property name.
		@param value property value; if null, the property is unset.
		@exception BadOperation the agent could not be found in local agency.
	*/
	public void setLocalAgentProperty(Name agent, String property, String value)
		throws BadOperation
	{
		AgentInfo entry = (AgentInfo)agentsByName.get(agent);
		if (entry != null)
		{
			setLocalAgentProperty(entry.getAgent(), property, value);
		}
		else
		{
			throw new BadOperation(BadOperation.UNKNOWNAGENT, "Exception in setLocalAgentProperty(): unknown agent " + agent + ".");
		}
	}


	/**
		Sets, changes or unsets an agent property.
		@param agent the agent's Java object reference.
		@param property property name.
		@param value property value; if null, the property is unset.
		@throws BadOperation the agent could not be found in local agency.
	*/
	public void setLocalAgentProperty(MobileObject agent, String property, String value)
		throws BadOperation
	{
        try
		{
			shutdownLock.in();
		}
		catch (ShutdownLock.ShutdownException e)
		{
			throw Constants.agencyIsClosed;
		}
		try
		{
			synchronized(agent)
			{
				AgentInfo entry = (AgentInfo)agentsByRef.get(agent);
				if (entry == null)
				{
					throw new BadOperation(BadOperation.UNKNOWNAGENT, "Exception in setLocalAgentProperty(): unknown agent.");
				}
				Properties props = entry.getProperties();
				String oldValue;
				if (value == null)
				{
					oldValue = (String)props.remove(value);
					if (oldValue == null)
					{
						return;
					}
				}
				else
				{
					oldValue = (String)props.setProperty(property, value);
				}
				if (my_finder != null)
				{
					try
					{
						my_finder.register_agent(
							entry.getName().getmafname(),
							my_location.toString(),
							makeAgentProfile(props));
					}
					catch (Exception ex)
					{
						BadOperation exception = null;
						System.err.println("Exception in setLocalAgentProperty() for agent " + entry.getName() + ": " + ex);
						if (ex instanceof SystemException &&
							((SystemException)ex).completed.value() != CompletionStatus._COMPLETED_YES)
						{
							exception = new BadOperation(
								BadOperation.INFRASTRUCTURE,
								"Exception with MAFFinder in property change operation for agent " + entry.getName(),
								ex);
						}
						else if (ex instanceof NameInvalid)
						{
							exception = new BadOperation(
								BadOperation.INVALIDNAME,
								"Exception in property change operation for agent " + entry.getName() + ": invalid name.",
								ex);
						}
						if (exception != null)
						{
							if (oldValue == null)
							{
								props.remove(property);
							}
							else
							{
								props.setProperty(property, oldValue);
							}
							throw exception;
						}
					}
				}
			}
        }
		finally
		{
			shutdownLock.out();
		}
	}


	/**
		@return true if the designated agent is currently active, false if
		the agent is suspended.
		@param agent wrapper to the agent's MAF name.
		@throws BadOperation the agent could not be found in local agency.
	*/
	public boolean localAgentIsRunning(Name agent)
		throws BadOperation
	{
		AgentInfo entry = (AgentInfo)agentsByName.get(agent);
		if (entry != null)
		{
			return localAgentIsRunning(entry.getAgent());
		}
		else
		{
			throw new BadOperation(BadOperation.UNKNOWNAGENT, "Can't state activity status of unknown agent " + agent + ".");
		}
	}


	/**
		@return true if the designated agent is currently active, false if
		the agent is suspended.
		@param agent the agent's Java object reference.
		@throws BadOperation the agent could not be found in local agency.
	*/
	public boolean localAgentIsRunning(MobileObject agent)
		throws BadOperation
	{
        try
		{
			shutdownLock.in();
		}
		catch (ShutdownLock.ShutdownException e)
		{
			throw Constants.agencyIsClosed;
		}
		try
		{
			synchronized(agent)
			{
				AgentInfo entry = (AgentInfo)agentsByRef.get(agent);
				if (entry != null)
				{
					return entry.isRunning();
				}
				else
				{
					throw new BadOperation(BadOperation.UNKNOWNAGENT, "Can't state activity status of unknown agent.");
				}
			}
		}
        finally
		{
			shutdownLock.out();
		}
	}


	/**
		@return the ORB object used by current agency.
	*/
	public ORB getORB()
	{
		return my_orb;
	}


	/**
	 * Add an agent to the agency's internal tables.
	 * @param agent the agent to be added
	 * @param the name of the agent
	 * @param entry information record on the agent
	 */
	private void registerAgent(MobileObject agent, Name name, AgentInfo entry)
	{
		synchronized (agentsByRef)
		{
			agentsByName.put(name, entry);
			agentsByRef.put(agent, entry);
		}
	}


	/**
	 * Remove an agent from the agency's internal tables.
	 * @param agent the agent to remove
	 * @param the name of the agent
	 */
	private void unregisterAgent(MobileObject agent, Name name)
	{
		synchronized(agentsByRef)
		{
			agentsByName.remove(name);
			agentsByRef.remove(agent);
		}
	}


	/**
		The agency may invoke System.exit() to make the JVM exit
		when the agency is terminated.
		@param mode true enables the automatic exit on agency termination, false disables.
		@param delay delay in ms before calling System.exit(), after all callbacks
		(on agents, on agency personality) have returned; a negative value results in
		(1) dedicated system property, or (2) previous value set, or (3) default value to be applied.
		@see Constants
	*/
	public void exitOnTerminate(boolean mode, long delay)
	{
		standalone = mode;
		if (standalone && delay >= 0)
		{
			System.setProperty(Constants.agencyExitDelayProp, String.valueOf(delay));
		}
	}


	/**
	 * Internal class to launch a delayed exit of process
	 */
	class ExitDelay extends Thread
	{
		long delay;

		public ExitDelay(long ms_delay)
		{
			delay = ms_delay;
			start();
		}

		public void run()
		{
			try
			{
				sleep(delay);
			}
			catch (InterruptedException ex)
			{
				System.err.println(ex.toString());
			}
			System.exit(0);
		}
	}


	///////////////////////////////////////////////////
	// implementation of interface AgencyPersonality //
	///////////////////////////////////////////////////


	/**
	 * Call afterBirth() on agent.
	*/
	public void afterBirth(MobileObject agent, AgentSystem agency, AgentInfo entry, Object arguments)
		throws BadOperation
	{
		agent.afterBirth(agency, entry, arguments);
	}


	/**
	 * Call beforeMove() on agent.
	 */
	public void beforeMove(MobileObject agent, Location location, String place)
		throws BadOperation
	{
		agent.beforeMove(location, place);
	}


	/**
	 * Call afterMove() on agent.
	 */
	public void afterMove(MobileObject agent, AgentSystem agency, Location location, String place)
		throws BadOperation
	{
		agent.afterMove(agency, location, place);
	}


	/**
	 * Call afterMoveFailed on agent.
	 */
	public void afterMoveFailed(
		MobileObject agent,
		Location agency,
		String place,
		int reason,
		String message)
	{
		agent.afterMoveFailed(agency, place, reason, message);
	}


	/**
	 * Call beforeDeath() on agent
	 */
	public void beforeDeath(MobileObject agent)
	{
		agent.beforeDeath();
	}


	/**
	 * Call beforeSuspend() on agent
	 */
	public void beforeSuspend(MobileObject agent)
		throws BadOperation
	{
		agent.beforeSuspend();
	}


	/**
	 * Call beforeResume() on agent
	 */
	public void beforeResume(MobileObject agent)
		throws BadOperation
	{
		agent.beforeResume();
	}


	/**
	 * Call beforeShutdown() on agent
	 */
	public void beforeShutdown(MobileObject agent)
	{
		agent.beforeShutdown();
	}


	/**
	 * Nothing special to do
	 */
	public void agencyShutdown()
	{
	}
}
