/*
* MobiliTools: an implementation of the Object Management Group's
* Mobile Agent Facility specification.
* Copyright (C) 2003 France Telecom R&D
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
* MobiliTools $Name:  $
*
* Contact: mobilitools-smi@lists.debian-sf.objectweb.org
*
* Authors: Bruno Dillenseger
*/


package org.objectweb.mobilitools.smi.api;

import org.omg.CfMAF.AgentSystemInfo;

/**
 * MobiliTools $Name:  $, $Id: AgencyPersonality.java,v 1.1.1.1 2003/03/28 14:48:06 dillense Exp $
 * <P>
 * Interface that must be implemented by an agency personality
 * to give hooks to SMI generic Agency. Most of this interface
 * consists in intercepting MobileObject life cycle.
 * @see Agency
 * @see MobileObject
*/
public interface AgencyPersonality
{
	/**
		Called on agent creation.
		@param agent target agent.
		@param agency reference to the hosting agency.
		@param entry a structure of information about the newly created agent.
		@param arguments arguments for agent initialization.
		@throws BadOperation shall be thrown if the agent refuses to settle.
		Identifier BadOperation.REJECTED must be set for this exception.
		Note that the agent creation is canceled,
		and that <code>beforeDeath()</code> is not called.
	*/
	public void afterBirth(MobileObject agent, AgentSystem agency, AgentInfo entry, Object arguments)
		throws BadOperation;

	/**
		Called before before moving.
		@param agent target agent
		@param location location of destination agency
		@param place name of target place
		@throws BadOperation thrown to reject the move.
		Identifier BadOperation.REJECTED must be set for this exception.
		The agent move is simply canceled and the thrown exception is
		re-thrown by <code>moveAgent()</code> (note that <code>afterMoveFailed()</code>
		is not called).
	*/
	public void beforeMove(MobileObject agent, Location location, String place)
		throws BadOperation;

	/**
		Called on reinstalling in the target agency after a move.
		@param agent target agent
		@param agency reference to the new hosting agency
		@param location location of new hosting agency
		@param place the name of the new residing place
		@throws BadOperation thrown if the agent can not be
		correctly reinstalled in the new agency.
		Identifier BadOperation.REJECTED must be set for this exception.
		The agent stays in the source agency, <code>afterMoveFailed()</code> is invoked,
		and <code>moveAgent()</code> throws an exception.
	*/
	public void afterMove(MobileObject agent, AgentSystem agency, Location location, String place)
		throws BadOperation;

	/**
		Called on move failure. (note that throwing an exception in
		<code>beforeMove()</code> does not result in
		<code>afterMoveFailed()</code> being called)
		@param agent target agent
		@param agency location of the target agency
		@param place the target place name
		@param reason problem identifier
		@param message textual explanation
	*/
	public void afterMoveFailed(
		MobileObject agent,
		Location agency,
		String place,
		int reason,
		String message);

	/**
		Called before discarding the agent.
		(note that an exception in <code>afterBirth()</code> does not
		result in <code>beforeDeath()</code> being called)
		@param agent target agent
	*/
	public void beforeDeath(MobileObject agent);

	/**
		Called to suspend the agent activity (if any).
		@param agent target agent
		@throws BadOperation the agent activity was already suspended, or could not be suspended.
		Identifier BadOperation.REJECTED must be set for this exception. SMI core detects
		whether the agent is already suspended or not, and throws the corresponding exception if
		necessary, without invoking this callback.
	*/
	public void beforeSuspend(MobileObject agent)
		throws BadOperation;

	/**
		Called to resume the agent activity (if any).
		@param agent target agent
		@throws BadOperation the agent activity was already running, or could not be resumed.
		Identifier BadOperation.REJECTED must be set for this exception. SMI core detects
		whether the agent is already running or not, and throws the corresponding exception if
		necessary, without invoking this callback.
	*/
	public void beforeResume(MobileObject agent)
		throws BadOperation;

	/**
		Called before the agent's hosting agency shutdown.
		If the agent is still residing in the shutting agency
		after this callback, it is terminated and callback
		beforeDeath() is invoked (an agent can't survive the
		shutdown of its hosting agency).
		@param agent target agent
	*/
	public void beforeShutdown(MobileObject agent);

	/**
	 * Called before agency shutdown (once all agents have been warned and terminated).
	 */
	public void agencyShutdown();
}
