/*
* MobiliTools: an implementation of the Object Management Group's
* Mobile Agent Facility specification.
* Copyright (C) 2003 France Telecom R&D
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
* MobiliTools $Name:  $
*
* Contact: mobilitools-smi@lists.debian-sf.objectweb.org
*
* Authors: Bruno Dillenseger
*/


package org.objectweb.mobilitools.smi.api;


import java.io.*;


/**
 * MobiliTools $Name:  $, $Id: BadOperation.java,v 1.1.1.1 2003/03/28 14:48:05 dillense Exp $
 * <P>
 * Exception indicating that a request has failed.
*/
public class BadOperation extends Exception
{
	/** unspecified failure = 0 */
	static public final int OTHER = 0;
	/** the target agent rejected the operation = 1 */
	static public final int REJECTED = 1;
	/** infrastructure (network, CORBA, naming service...) problem = 2 */
	static public final int INFRASTRUCTURE = 2;
	/** destination agency for agent move is unknown = 3*/
	static public final int DESTINATION = 3;
	/** agent class fetch problem = 4 */
	static public final int CLASSFAULT = 4;
	/** the target agent is unknown = 5 */
	static public final int UNKNOWNAGENT = 5;
	/** agent or argument object de/serialization failure = 6 */
	static public final int SERIALIZATION = 6;
	/** agent is active after a suspend or resume request failure = 7 */
	static public final int RUNNING = 7;
	/** agent is suspended after a suspend or resume request failure = 8 */
	static public final int SUSPENDED = 8;
	/** agent class could not be instantiated (check access, security, abstract class, static constructor...) = 9 */
	static public final int INSTANTIATION = 9;
	/** agent type is not recognized = 10 */
	static public final int AGENTTYPE = 10;
	/** name is invalid (already exists? other?)  = 11 */
	static public final int INVALIDNAME = 11;
	/** unknown agency = 13 */
	static public final int UNKNOWNAGENCY = 13;
	/** a (temporary/normal or permanent/abnormal?) unconsistency happened, retry later = 14 */
	static public final int RETRY = 14;


	/** failure identification */
	protected int reason = OTHER;
	/** failure explanation message */
	protected String message = null;
	/** nested exception */
	protected Exception nested = null;


	/**
		Creates a new exception to signal any request failure.
		@param reason failure identification - should be one of the identifiers defined in this class.
		@param message failure explanation message.
	*/
	public BadOperation(int reason, String message)
	{
		this(reason, message, null);
	}


	/**
		Creates a new exception to signal any request failure.
		@param reason failure identification - should be one of the identifiers defined in this class.
		@param message failure explanation message.
		@param nested the actual exception this BadOperation springs from.
	*/
	public BadOperation(int reason, String message, Exception nested)
	{
		this.reason = reason;
		this.message = message;
		this.nested = nested;
	}


	/**
		@return failure reason identifer.
	*/
	public int getReason()
	{
		return reason;
	}


	/**
		@return failure explanation message.
	*/
	public String getMessage()
	{
		return message;
	}


	/**
		@return the nested exception (if any is defined), null otherwise.
	*/
	public Exception getNested()
	{
		return nested;
	}


	/**
		@return a message describing the exception.
	*/
	public String toString()
	{
		return "BadOperation exception: " + message + " (error code " + String.valueOf(reason) + (nested == null ? ")" : ", nested exception: " + nested + ")");
	}


	/**
		Prints the execution stack trace first, and then
		the nested exception stack trace if any.
	*/
	public void printAllStackTrace()
	{
		printAllStackTrace(System.err);
	}


	/**
		Prints the execution stack trace first, and then
		the nested exception stack trace if any.
		@param s where to print stack trace.
	*/
	public void printAllStackTrace(PrintStream s)
	{
		printStackTrace(s);
		if (nested != null)
		{
			s.println("NESTED EXCEPTION'S STACK TRACE BELOW:");
			nested.printStackTrace(s);
		}
	}


	/**
		Prints the execution stack trace first, and then
		the nested exception stack trace if any.
		@param w where to print stack trace.
	*/
	public void printAllStackTrace(PrintWriter w)
	{
		printStackTrace(w);
		if (nested != null)
		{
			w.println("NESTED EXCEPTION'S STACK TRACE BELOW:");
			nested.printStackTrace(w);
		}
	}
}
