/*
* MobiliTools: an implementation of the Object Management Group's
* Mobile Agent Facility specification.
* Copyright (C) 2003 France Telecom R&D
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
* MobiliTools $Name:  $
*
* Contact: mobilitools-smi@lists.debian-sf.objectweb.org
*
* Authors: Bruno Dillenseger
*/


package org.objectweb.mobilitools.smi.api;

import org.omg.CfMAF.AgentSystemInfo;
import org.omg.CfMAF.LanguageMap;
import org.omg.CORBA.Any;


/**
 * MobiliTools $Name:  $, $Id: Constants.java,v 1.1.1.1 2003/03/28 14:48:05 dillense Exp $
 * <P>
 * Set of constants for parameterizing SMI.
 * <UL>
 *   <LI>MAF-related identifiers for SMI agent system
 *   <LI>system properties for agency naming (re. CORBA naming service)
 * 	    <BLOCKQUOTE>
 *        Any SMI agency is automatically bound to a name in the CORBA naming service.
 *        This name is constructed according to the scheme:
 *        <BR><code>/region prefix/region name/agency prefix/agency name</code>
 *        <BR>Both prefixes have default values which may be overridden by setting
 *        dedicated system properties. (String'fied COSnaming names have the following format:
 *        <BR><code>/naming context1!kind1/.../object!kind</code>)
 *      </BLOCKQUOTE>
 * </UL>
*/
public abstract class Constants
{
	// identifiers for SMI agent systems

	/** agent system type identifier for SMI = 10 */
	public static final short TYPE_ID = 10;
	/** major version of SMI agent system = 1 */
	public static final short MAJOR_VERSION = 1;
	/** minor version of SMI agent system = 0 */
	public static final short MINOR_VERSION = 0;
	/** textual description of SMI agent system = "Simple MAF Implementation" */
	public static final String DESCRIPTION = "Simple MAF Implementation";


	// identifiers compliant with MAF specifications

	/** MAF-compliant language identifier for SMI agent system = 1 */
	public static final short LANGUAGE_ID = 1;
	/** MAF-compliant serialization identifier for SMI agent system = 1 */
	public static final short SERIALIZATION = 1;
	/** COS naming URI prefix = "CosNaming:" */
	public static final String COSNAMINGURI = "CosNaming:";


	// properties and default values for the agency's full name in the CORBA Naming Service

	/** property name for region naming prefix = "org.objectweb.mobilitools.smi.regionPrefix" */
	public static final String regionPrefixProp = "org.objectweb.mobilitools.smi.regionPrefix";
	/** default value for region naming prefix property = "MAF" */
	public static final String regionPrefixDefault = "MAF";
	/** property name for agency naming prefix = "org.objectweb.mobilitools.smi.agencyPrefix" */
	public static final String agencyPrefixProp = "org.objectweb.mobilitools.smi.agencyPrefix";
	/** default value for agency naming prefix = "agency" */
	public static final String agencyPrefixDefault = "agency";
	/** property name for finder naming prefix = "org.objectweb.mobilitools.smi.finderPrefix" */
	public static final String finderNameProp = "org.objectweb.mobilitools.smi.finderPrefix";
	/** default value for finder naming prefix = "finder" */
	public static final String finderNameDefault = "finder";


	// when an agency is remotely terminated, it must wait a while so that the CORBA call
	// has returned before exiting. (only applies when Agency standalone attribute in true)

	/** property name for setting exit delay when agencies are terminated = "org.objectweb.mobilitools.smi.agencyExitDelay" */
	public static final String agencyExitDelayProp = "org.objectweb.mobilitools.smi.agencyExitDelay";
	/** default exit delay in ms when agencies are terminated = "10000" */
	public static final String agencyExitDelayDefault = "10000";


	// properties and default values for classloader factory class

	/** property name for classloader factory = "org.objectweb.mobilitools.smi.classloader" */
	public static final String classLoaderProp = "org.objectweb.mobilitools.smi.classloader";
	/** default value for classloader factory = "org.objectweb.mobilitools.smi.lib.SMIClassLoader" */
	public static final String classLoaderDefault = "org.objectweb.mobilitools.smi.lib.SMIClassLoader";
	/** method name called on classloader factory is "getClassLoader" */
	public static final String classLoaderMethod = "getClassLoader";


	// exceptions

	/** exception raised by some Agency methods when the agency is shutdown */
	public static final RuntimeException agencyIsClosed =
		new RuntimeException("The agency is terminated.");

	/**
	 * @param identity the agent system's identity
	 * @param authority the agent system's authority
	 * @return a MAF agent system information record for an SMI agency
	 */
	public static final AgentSystemInfo makeAgentSystemInfo(byte[] authority, byte[] identity)
	{
		return new AgentSystemInfo(
			new org.omg.CfMAF.Name(
				authority,
				identity,
				TYPE_ID),
			TYPE_ID,
			new LanguageMap[] {
				new LanguageMap(
					LANGUAGE_ID,
					new short[] { SERIALIZATION })
					},
			DESCRIPTION,
			MAJOR_VERSION,
			MINOR_VERSION,
			new Any[0]);
	}


	/**
	 * Displays version information
	 */
	public static void main(String[] args)
	{
		System.out.println("MobiliTools version $Name:  $");
	}
};
