/*
* MobiliTools: an implementation of the Object Management Group's
* Mobile Agent Facility specification.
* Copyright (C) 2003 France Telecom R&D
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
* MobiliTools $Name:  $
*
* Contact: mobilitools-smi@lists.debian-sf.objectweb.org
*
* Authors: Bruno Dillenseger
*/


package org.objectweb.mobilitools.smi.api;


import java.io.Serializable;

/**
 * MobiliTools $Name:  $, $Id: Location.java,v 1.1.1.1 2003/03/28 14:48:06 dillense Exp $
 * <P>
 * SMI locations are based on a special naming scheme, based on CORBA naming service
 * (see COS naming URI in MAF specification):
 * CosNaming:region_prefix/region_name/agency_prefix/agency_name
 * e.g. CosNaming:MAF/my_region/agency/my_agency
 * Default prefixes are defined in Constants class, and can be overridden by
 * setting appropriate properties.
*/
public class Location implements Serializable
{
	static String regionPrefix = System.getProperty(Constants.regionPrefixProp, Constants.regionPrefixDefault);
	static String agencyPrefix = System.getProperty(Constants.agencyPrefixProp, Constants.agencyPrefixDefault);
	String my_region;
	String my_agency;
	String my_location;


	/**
		Creates a new Location object from the provided region name and agency name.
	*/
	public Location(String region, String agency)
	{
		my_region = region;
		my_agency = agency;
		my_location =
			Constants.COSNAMINGURI +
			regionPrefix +
			"/" +
			region +
			"/" +
			agencyPrefix +
			"/" +
			agency;
	}


	/**
		Creates a new Location object from the provided location string.
		String format should be: CosNaming:&lt;region_prefix&gt;/&lt;region_name&gt;/&lt;agency_prefix&gt;/&lt;agency_name&gt;,
		according to default prefixes defined in Constants class, or according to properties overriding
		these default prefixes.
		@see Constants
		@throws BadOperation the provided string format is not recognized as a location.
	*/
	public Location(String location)
		throws BadOperation
	{
		if (location == null)
		{
			throw new BadOperation(BadOperation.REJECTED, location + " is not a location string.");
		}
		my_location = location;
		if (! location.startsWith(Constants.COSNAMINGURI))
		{
			throw new BadOperation(BadOperation.REJECTED, location + " is not a location string.");
		}
		location = location.substring(Constants.COSNAMINGURI.length());
		if (! location.startsWith(regionPrefix + "/"))
		{
			throw new BadOperation(BadOperation.REJECTED, location + " is not a location string.");
		}
		if (regionPrefix.length() > 0)
		{
			location = location.substring(regionPrefix.length()+1);
		}
		int beginAgency = location.lastIndexOf("/");
		if (beginAgency == -1 || beginAgency == location.length()-1)
		{
			throw new BadOperation(BadOperation.REJECTED, location + " is not a location string.");
		}
		my_agency = location.substring(beginAgency + 1);
		try
		{
			if (agencyPrefix.length() > 0)
			{
				my_region = location.substring(0, beginAgency - agencyPrefix.length() - 1);
			}
			else
			{
				my_region = location.substring(0, beginAgency);
			}
		}
		catch (IndexOutOfBoundsException e)
		{
			throw new BadOperation(BadOperation.REJECTED, location + " is not a location string.");
		}
	}


	/**
		@return region name for this location.
	*/
	public String getRegion()
	{
		return my_region;
	}


	/**
	   @return agency name for this location.
	*/
	public String getAgency()
	{
		return my_agency;
	}


	public String toString()
	{
		return my_location;
	}


	public boolean equals(Object other)
	{
		if (other instanceof Location)
		{
			return
				my_region.equals(((Location)other).getRegion()) &&
				my_agency.equals(((Location)other).getAgency());
		}
		else
		{
			return false;
		}
	}


	/**
		@return location's mapping to a string'fied name in CORBA naming service.
	*/
	public String getCosNamingName()
	{
		return my_location.substring(Constants.COSNAMINGURI.length());
	}
}
