/*
* MobiliTools: an implementation of the Object Management Group's
* Mobile Agent Facility specification.
* Copyright (C) 2003 France Telecom R&D
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
* MobiliTools $Name:  $
*
* Contact: mobilitools-smi@lists.debian-sf.objectweb.org
*
* Authors: Bruno Dillenseger
*/


package org.objectweb.mobilitools.smi.goodies;


import javax.swing.*;
import javax.swing.table.AbstractTableModel;
import javax.swing.event.TableModelEvent;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Observer;
import java.util.Observable;
import java.util.StringTokenizer;
import org.objectweb.mobilitools.smi.*;
import org.objectweb.mobilitools.smi.api.*;
import org.objectweb.mobilitools.util.gui.*;
import org.omg.CORBA.ORB;
import org.omg.CORBA.Any;
import org.omg.CfMAF.AgentProfile;
import org.omg.CfMAF.AgentSystemInfo;


/**
 * MobiliTools $Name:  $, $Id: FinderGUI.java,v 1.1.1.1 2003/03/28 14:48:06 dillense Exp $
 * <P>
 * GUI for SMI Finder. May be run as a standalone programme which launches a new finder
 * and its GUI, or may be used to create a GUI for an existing finder.
*/
public class FinderGUI
	extends JFrame
	implements Observer
{
	static boolean standalone = false;
	static String[] columnNames = {	"Name", "Location" };


	/**
		Creates a new finder with a GUI.
		@param args the last argument must specify the region name.
	*/
	static public void main(String[] args)
	{
		ORB orb = ORB.init(args, System.getProperties());
		standalone = true;
		new FinderGUI(new Finder(args[args.length-1], orb));
		orb.run();
	}


	Finder my_finder;
	Object[][] data;
	TableSorter tableSorter;
	MyTableModel tableModel;
	JTable table;
	RegionManager region;


	/**
		Opens a GUI for the specified Finder.
		@param finder the SMI finder to browse.
	*/
	public FinderGUI(Finder finder)
	{
		super("SMI Finder for region " + finder.getRegion());
		my_finder = finder;
		try
		{
			region = new RegionManager(finder.getORB());
		}
		catch (BadOperation e)
		{
			System.err.println(e.toString());
			region = null;
		}
		data = new String[0][2];
		finder.addObserver(this);
		addWindowListener(new OnWindowClosing());

        tableModel = new MyTableModel();
        tableSorter = new TableSorter(tableModel);
        table = new JTable(tableSorter);
		table.addMouseListener(new OnMouseClick());
        tableSorter.addMouseListenerToHeaderInTable(table);
		//      table.setPreferredScrollableViewportSize(new Dimension(500, 70));

        //Create the scroll pane and add the table to it.
        JScrollPane scrollPane = new JScrollPane(table);

        //Add the scroll pane to this window.
        getContentPane().add(scrollPane, BorderLayout.CENTER);

		pack();
        setVisible(true);
	}


	/**
		Terminates the GUI.
		If FinderGUI was launched as a standalone programme, the attached MAFFinder is terminated,
		and the programme exits.
	*/
	public void quit()
	{
		dispose();
		if (standalone)
		{
			System.exit(my_finder.exit());
		}
	}


	//////////////////////////////////////////
	// Implementation of interface Observer //
	//////////////////////////////////////////


	public void update(Observable o, java.lang.Object arg)
	{
		FinderEntry[] entries = (FinderEntry[])arg;
		data = new Object[entries.length][];
		for (int i=0 ; i<data.length ; ++i)
		{
			data[i] = new Object[] { entries[i], entries[i].getLocation() };
		}
		tableSorter.tableChanged(new TableModelEvent(tableModel));
		table.clearSelection();
	}


	///////////////////////////////////////
	// refinement of class WindowAdapter //
	///////////////////////////////////////


	class OnWindowClosing extends WindowAdapter
	{
		public void windowClosing(WindowEvent e)
		{
			FinderGUI.this.quit();
		}
	}


	/////////////////////////////////////////////////////////
	// inner class for managing double-clicks on the table //
	/////////////////////////////////////////////////////////


	class OnMouseClick extends MouseAdapter
	{
		public void mouseClicked(MouseEvent e)
		{
			if (e.getClickCount() == 2)
			{
				FinderEntry selection = (FinderEntry)tableSorter.getValueAt(FinderGUI.this.table.rowAtPoint(e.getPoint()), 0);
				if (selection.isAgent())
				{
					new AgentWindow(FinderGUI.this, selection);
				}
				else
				{
					new AgentSystemWindow(FinderGUI.this, selection);
				}
			}
		}
	}


	////////////////////////////////////////////////////
	// inner class for giving the table a table model //
	////////////////////////////////////////////////////


    class MyTableModel extends AbstractTableModel {
        public int getColumnCount() {
            return FinderGUI.columnNames.length;
        }

        public int getRowCount() {
            return FinderGUI.this.data.length;
        }

        public String getColumnName(int col) {
            return FinderGUI.columnNames[col];
        }

        public Object getValueAt(int row, int col) {
            return FinderGUI.this.data[row][col];
        }
	}


	///////////////////////////////////////////////////
	// inner class for viewing and managing an agent //
	///////////////////////////////////////////////////


	class AgentWindow extends JDialog implements ActionListener, FormListener
	{
		FinderEntry my_agent;
		JButton moveBtn, suspendBtn, resumeBtn, terminateBtn;


		AgentWindow(Frame frame, FinderEntry agent)
		{
			super(frame, agent.toString());
			my_agent = agent;
			boolean isRunning;
			String infoStr =
				"Name: " + agent.getName().identity() + "\n" +
				"Authority: " + agent.getName().authority() + "\n" +
				"Type: " + String.valueOf(agent.getName().type()) + "\n" +
				"Location: " + agent.getLocation() + "\n" +
				"Activity: ";
			Container pane = this.getContentPane();
			JPanel buttonPnl = new JPanel();
			buttonPnl.setLayout(new FlowLayout());
			buttonPnl.add(moveBtn = new JButton("move"));
			moveBtn.addActionListener(this);
			try
			{
				isRunning = region.isRunning(new Location(agent.getLocation()), agent.getName());
				if (isRunning)
				{
					buttonPnl.add(suspendBtn = new JButton("suspend"));
					suspendBtn.addActionListener(this);
					resumeBtn = null;
					infoStr += "running\n";
				}
				else
				{
					buttonPnl.add(resumeBtn = new JButton("resume"));
					resumeBtn.addActionListener(this);
					suspendBtn = null;
					infoStr += "suspended\n";
				}
			}
			catch (Exception e)
			{
				System.err.println("Warning: can't check " + agent + "'s activity status:\n" + e);
				infoStr += "unknown\n";
			}
			StringWriter strWriter = new StringWriter();
			agent.getProperties().list(new PrintWriter(strWriter));
			infoStr += strWriter.toString();
			buttonPnl.add(terminateBtn = new JButton("terminate"));
			terminateBtn.addActionListener(this);
			pane.add(buttonPnl, BorderLayout.SOUTH);
			JTextArea infoTxt = new JTextArea(infoStr);
			infoTxt.setEditable(false);
			pane.add(new JScrollPane(infoTxt), BorderLayout.CENTER);
			this.pack();
			this.setVisible(true);
		}


		public void actionPerformed(ActionEvent ev)
		{
			if (ev.getSource() == moveBtn)
			{
				new FormFrame(
					"Move " + my_agent,
					"Enter destination for " + my_agent,
					new String[] { "region", "agency", "place" },
					new String[] { "move", "stay" }
					).run(this);
			}
			else if (ev.getSource() == resumeBtn)
			{
				try
				{
					region.resumeAgent(new Location(my_agent.getLocation()), my_agent.getName());
				}
				catch (Exception ex)
				{
					new AlertWindow(FinderGUI.this, "Can't resume " + my_agent, ex.toString(), "OK");
				}
				this.dispose();
			}
			else if (ev.getSource() == suspendBtn)
			{
				try
				{
					region.suspendAgent(new Location(my_agent.getLocation()), my_agent.getName());
				}
				catch (Exception ex)
				{
					new AlertWindow(FinderGUI.this, "Can't suspend " + my_agent, ex.toString(), "OK");
				}
				this.dispose();
			}
			else if (ev.getSource() == terminateBtn)
			{
				try
				{
					region.terminateAgent(new Location(my_agent.getLocation()), my_agent.getName());
				}
				catch (Exception ex)
				{
					new AlertWindow(FinderGUI.this, "Can't terminate " + my_agent, ex.toString(), "OK");
				}
				this.dispose();
			}
		}


		public void formResult(String[] result)
		{
			if (result[result.length-1].equals("move"))
			{
				try
				{
					region.moveAgent(
						new Location(my_agent.getLocation()),
						my_agent.getName(),
						new Location(result[0], result[1]),
						result[2]);
				}
				catch (Exception e)
				{
					new AlertWindow(FinderGUI.this, "Can't move " + my_agent, e.toString(), "OK");
				}
			}
			this.dispose();
		}
	}


	//////////////////////////////////////////////////////////
	// inner class for viewing and managing an agent system //
	//////////////////////////////////////////////////////////


	class AgentSystemWindow extends JDialog implements ActionListener, FormListener
	{
		FinderEntry my_agency;
		JButton createBtn, terminateBtn;


		AgentSystemWindow(Frame frame, FinderEntry agency)
		{
			super(frame, agency.toString());
			my_agency = agency;
			Container pane = this.getContentPane();
			StringWriter strWriter = new StringWriter();
			agency.getProperties().list(new PrintWriter(strWriter));
			JTextArea infoTxt = new JTextArea(
				"Name: " + agency.getName().identity() + "\n" +
				"Authority: " + agency.getName().authority() + "\n" +
				"Type: " + String.valueOf(agency.getName().type()) + "\n" +
				"Location: " + agency.getLocation() + "\n" +
				strWriter.toString());
			infoTxt.setEditable(false);
			pane.add(new JScrollPane(infoTxt), BorderLayout.CENTER);
			JPanel buttonPnl = new JPanel();
			buttonPnl.setLayout(new FlowLayout());
			buttonPnl.add(createBtn = new JButton("new agent"));
			createBtn.addActionListener(this);
			buttonPnl.add(terminateBtn = new JButton("terminate"));
			terminateBtn.addActionListener(this);
			pane.add(buttonPnl, BorderLayout.SOUTH);
			this.pack();
			this.setVisible(true);
		}


		public void actionPerformed(ActionEvent ev)
		{

			if (ev.getSource() == createBtn)
			{
				new FormFrame(
					"New agent in " + my_agency,
					"Enter new agent parameters",
					new String[] {
						"class name",
						"agent name",
						"place name",
						"code base",
						"arguments" },
					new String[] { "create", "cancel" }
					).run(this);
			}
			else if (ev.getSource() == terminateBtn)
			{
				try
				{
					region.terminateAgency(new Location(my_agency.getLocation()));
				}
				catch (Exception ex)
				{
					new AlertWindow(FinderGUI.this, "Can't terminate " + my_agency, ex.toString(), "OK");
				}
				this.dispose();
			}
		}


		public void formResult(String[] result)
		{
			if (result[result.length-1].equals("create"))
			{
				StringTokenizer parser = new StringTokenizer(result[4], " ");
				String[] args = new String[parser.countTokens()];
				for (int i=0 ; i<args.length ; ++i)
				{
					args[i] = parser.nextToken();
				}
				AgentSystemInfo agencyInfo = (AgentSystemInfo)my_agency.getProfile();
				org.objectweb.mobilitools.smi.api.Name name = new org.objectweb.mobilitools.smi.api.Name(
					System.getProperty("user.name"),
					result[1],
					agencyInfo.agent_system_type);

				try
				{
					region.createAgent(
						new Location(my_agency.getLocation()),
						result[0],
						result[3],
						name,
						result[2],
						new AgentProfile(
							agencyInfo.language_maps[0].language_id,
							agencyInfo.agent_system_type,
							agencyInfo.agent_system_description,
							agencyInfo.major_version,
							agencyInfo.minor_version,
							agencyInfo.language_maps[0].serializations[0],
							new Any[0]),
						null,
						args);
				}
				catch (Exception ex)
				{
					new AlertWindow(
						FinderGUI.this,
						"Agent creation failed",
						name + " could not be created in " + my_agency + ":\n" + ex,
						"OK");
				}
			}
			this.dispose();
		}
	}
}
