/*
* MobiliTools: an implementation of the Object Management Group's
* Mobile Agent Facility specification.
* Copyright (C) 2003 France Telecom R&D
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
* MobiliTools $Name:  $
*
* Contact: mobilitools-smi@lists.debian-sf.objectweb.org
*
* Authors: Bruno Dillenseger
*/


package org.objectweb.mobilitools.smi.goodies;


import java.util.*;
import java.io.*;
import org.omg.CfMAF.*;
import org.omg.CORBA.ORB;
import org.omg.CORBA.SystemException;
import org.omg.CORBA.Any;
import org.omg.CORBA.BAD_PARAM;
import org.objectweb.mobilitools.util.corba.NameService;
import org.objectweb.mobilitools.util.corba.NameServiceException;
import org.objectweb.mobilitools.util.corba.CORBA;
import org.objectweb.mobilitools.smi.*;
import org.objectweb.mobilitools.smi.api.*;
import org.objectweb.mobilitools.smi.idl.*;
import mparser.*;


/**
 * MobiliTools $Name:  $, $Id: RegionManager.java,v 1.1.1.1 2003/03/28 14:48:06 dillense Exp $
 * <P>
 * A RegionManager instance makes it possible to deploy and manage agents,
 * as well as lookup agents and agent systems, through the MAF infrastructure
 * (i.e. implementations of MAFAgentSystem and MAFFinder interfaces).
 * Localization of MAFAgentSystem and MAFFinder servers is based on the
 * CORBA naming service.
 * If one naming service is running for each region, then regions must be
 * interconnected by federating the naming services in the naming context
 * of region name bindings.
 * <BR>
 * RegionManager can also be run as a standalone programme in order to create
 * a set of agents in various existing agencies, by specifying a property file
 * or XML file where all necessary definitions will be read from.
*/
public class RegionManager
{
	NameService my_ns;
	ORB my_orb;

	/**
		Creates/deploys agents in existing agencies, from definitions
		read from a property file or an XML file.
		@param args shall contain ORB-specific switches first, and
		then either <code>-xml xml_file</code> or <code>-prop property_file</code>
		switch.
		@see #deployFromPropFile(String)
		@see #deployFromXMLFile(String)
	*/
	static public void main(String[] args)
	{
		try
		{
			RegionManager manager = new RegionManager(ORB.init(args, System.getProperties()));
			Properties props = new Properties();
			if (args[args.length-2].equalsIgnoreCase("-prop"))
			{
				manager.deployFromPropFile(args[args.length-1]);
			}
			else if(args[args.length-2].equalsIgnoreCase("-xml"))
			{
				manager.deployFromXMLFile(args[args.length-1]);
			}
			else
			{
				System.out.println("usage: [ORB-specific args...] [-prop file] [-xml file]");
			}
		}
		catch (Exception ex)
		{
			System.err.println(ex.toString());
		}
	}


	/**
		Parses the specified XML file in order to make the specified agent creations.
		This file should contain DEPLOY tags, embracing AGENT tags. If both tags,
		attributes name, region, agency, place, class, codebase, and arguments
		may be defined. When an attribute is defined in DEPLOY tag, then its value
		is the default value for every enclosed AGENT tag.
		@param file the name of the XML file to be parsed.
		@throws Exception a variety of exceptions if a parsing error occurs,
		because of an invalid file format (check the XML content)
	*/
	public void deployFromXMLFile(String file)
		throws Exception
	{
		WMparser parser = new WMparser(file);
		ItemParsed item;
		Properties props = null;
		int agentCount = 0;
		short status = 0;

		while (parser.hasMoreItems())
		{
			item = parser.getNextItem();
			switch (status)
			{
			case 0: // looking for <DEPLOY> tag
				if ((item.getType() == ItemParsed.ITEM_TYPE_ELEMT) && item.getName().equalsIgnoreCase("deploy"))
				{
					props = new Properties();
					agentCount = 0;
					status = 1;
				}
				else if (item.getType() == ItemParsed.ITEM_TYPE_TEXTE)
				{
					System.out.println(item.getValue());
				}
				else
				{
					System.out.println("XML deployment warning: ignoring \"" + item.getName() + "\"");
				}
				break;
			case 1: // looking for <DEPLOY> attributes, or <AGENT> tag, or <DEPLOY> tag
				if (item.getType() == ItemParsed.ITEM_TYPE_ATTRB)
				{
					props.setProperty(item.getName().substring("deploy.".length()), item.getValue());
				}
				else if ((item.getType() == ItemParsed.ITEM_TYPE_ELEMT) && item.getName().equalsIgnoreCase("deploy.agent"))
				{
					++agentCount;
					status = 2;
				}
				else if ((item.getType() == ItemParsed.ITEM_TYPE_ELEMT) && item.getName().equalsIgnoreCase("deploy"))
				{
					props = new Properties();
					agentCount = 0;
					status = 1;
				}
				else if (item.getType() == ItemParsed.ITEM_TYPE_TEXTE)
				{
					System.out.println(item.getValue());
				}
				else
				{
					System.out.println("XML deployment warning: ignoring \"" + item.getName() + "\"");
				}
				break;
			case 2: // looking for <AGENT> attributes, or <AGENT> tag, or <DEPLOY> tag
				if (item.getType() == ItemParsed.ITEM_TYPE_ATTRB)
				{
					props.setProperty(
						item.getName().substring("deploy.agent.".length()) + "." + String.valueOf(agentCount),
						item.getValue());
				}
				else if ((item.getType() == ItemParsed.ITEM_TYPE_ELEMT) && item.getName().equalsIgnoreCase("deploy.agent"))
				{
					++agentCount;
				}
				else if ((item.getType() == ItemParsed.ITEM_TYPE_ELEMT) && item.getName().equalsIgnoreCase("deploy"))
				{
					deploy(props);
					props = new Properties();
					agentCount = 0;
					status = 1;
				}
				else if (item.getType() == ItemParsed.ITEM_TYPE_TEXTE)
				{
					System.out.println(item.getValue());
				}
				else
				{
					System.out.println("XML deployment warning: ignoring \"" + item.getName() + "\"");
				}
				break;
			}
		}
		if (props != null)
		{
			deploy(props);
		}
	}


	/**
		Parses the specified property file in order to make the specified agent creations.
		@param file the name of the property file to be parsed.
		@throws IOException if the specified file could not be read.
		@see #deploy(Properties)
	*/
	public void deployFromPropFile(String file)
		throws IOException
	{
		Properties props = new Properties();
		props.load(new FileInputStream(new File(file)));
		deploy(props);
	}


	/**
		Performs agent creations specified by the provided properties.
		Properties are:
		<UL>
			<LI>name.<I>n</I> for <I>n</I>th agent name
			<LI>class.<I>n</I> for class name of <I>n</I>th agent
			<LI>codebase.<I>n</I> for codebase of <I>n</I>th agent
			<LI>arguments.<I>n</I> for <I>n</I>th agent's arguments
			<LI>region.<I>n</I> for region name of <I>n</I>th agent
			<LI>agency.<I>n</I> for agency name of <I>n</I>th agent
			<LI>place.<I>n</I> for place name of <I>n</I>th agent
		</UL>
		Where <I>n</I> increments from 1 and stops as soon as property name.<I>n</I> is undefined.
		Any of these properties but name.<I>n</I> may be given a default
		value for every agent by omitting ".<I>n</I>" suffix.
	*/
	public void deploy(Properties props)
	{
		String agentname = null;
		String classname = null;
		String codebase = null;
		String place = null;
		String argumentStr = null;
		String agency = null;
		String region = null;
		String suffix = null;
		org.omg.CfMAF.Name mafName = null;
		Object[] argument = null;
		Location location = null;

		try
		{
			StringTokenizer parser;
			AgentSystemInfo agencyInfo;
			for (int i=1 ; (agentname = props.getProperty("name." + (suffix = String.valueOf(i)))) != null ; ++i)
			{
				if ((classname = props.getProperty("class." + suffix)) == null)
				{
					classname = props.getProperty("class");
				}
				if ((place = props.getProperty("place." + suffix)) == null)
				{
					place = props.getProperty("place");
				}
				if ((codebase = props.getProperty("codebase." + suffix)) == null)
				{
					codebase = props.getProperty("codebase");
				}
				if ((argumentStr = props.getProperty("arguments." + suffix)) == null)
				{
					argumentStr = props.getProperty("arguments");
				}
				if ((agency = props.getProperty("agency." + suffix)) == null)
				{
					agency = props.getProperty("agency");
				}
				if ((region = props.getProperty("region." + suffix)) == null)
				{
					region = props.getProperty("region");
				}
				if (argumentStr == null)
				{
					argument = new Object[0];
				}
				else
				{
					parser = new StringTokenizer(argumentStr);
					argument = new String[parser.countTokens()];
					for (int j=0 ; j<parser.countTokens() ; ++j)
					{
						argument[j] = parser.nextToken();
					}
				}
				location = new Location(region, agency);
				agencyInfo = resolveAgentSystem(location).get_agent_system_info();
				mafName = new org.omg.CfMAF.Name(System.getProperty("user.name").getBytes(), agentname.getBytes(), agencyInfo.agent_system_type);
				System.out.println(
					"Creating agent " + new org.objectweb.mobilitools.smi.api.Name(mafName) + " of class " + classname +
					"\n- codebase " + codebase +
					"\n- place " + place + " at location " + location +
					"\n- with " + (argumentStr == null ? "no argument." : "arguments " + argumentStr));
				createAgent(
					location,
					classname,
					codebase,
					new org.objectweb.mobilitools.smi.api.Name(mafName),
					place,
					new AgentProfile(
						Constants.LANGUAGE_ID,
						agencyInfo.agent_system_type,
						agencyInfo.agent_system_description,
						agencyInfo.major_version,
						agencyInfo.minor_version,
						Constants.SERIALIZATION,
						new Any[0]),
					null,
					argument);
			}
		}
		catch (Exception ex)
		{
			ex.printStackTrace();
		}
	}


	public RegionManager(ORB orb)
		throws BadOperation
	{
		my_orb = orb;
		try
		{
			my_ns = new NameService(orb);
		}
		catch (NameServiceException e)
		{
			System.err.println("A RegionManager instance cannot proceed without COS naming service: " + e);
			throw new BadOperation(BadOperation.INFRASTRUCTURE, "Cannot create RegionManager instance: COS naming service not found.");
		}
	}


	public RegionManager()
		throws BadOperation
	{
		this(ORB.init(new String[0], System.getProperties()));
	}


	/**
		Creates an agent in the named agency of the specified region.
		The specified agent system is its own class provider.
		@return the actual name of the new agent, or null if the agent
		creation failed because the location could not be resolved.
		@param region region name
		@param agency name wrapper for agency MAF name
		@param classname the agent main class name.
		@param codebase the codebase for loading the agent classes
		(interpreted by the specified agent system).
		@param name the name of the new agent.
		@param place the name of place where the agent will be located.
		@param profile new agent profile. The agent type must be compatible
		with the agent system type of the target agent system.
		@param agent_data serializable object that will be passed as an argument
		to the afterBirth() agent method, unless the arguments parameter is not null.
		@param arguments array of arguments that will be passed as an argument
		to the afterBirth() agent method. Arguments must be Strings or object wrappers
		for Java primitive types. If this parameter is null, then the agent_data
		parameter is considered instead.
		@throws BadOperation agent creation failed.
		@see MobileObject#afterBirth(AgentSystem, AgentInfo, Object)
	*/
	public org.objectweb.mobilitools.smi.api.Name createAgent(
		String region,
		org.objectweb.mobilitools.smi.api.Name agency,
		String classname,
		String codebase,
		org.objectweb.mobilitools.smi.api.Name name,
		String place,
		AgentProfile profile,
		Serializable agent_data,
		Object[] arguments)
		throws BadOperation
	{
		return createAgent(
			findAgency(region, agency),
			classname,
			codebase,
			name,
			place,
			profile,
			agent_data,
			arguments);
	}


	/**
		Creates an agent at the specified agent system location.
		The specified agent system is its own class provider.
		@return the actual name of the new agent, or null if the agent
		creation failed because the location could not be resolved.
		@param location agent system location where the agent must me created.
		@param classname the agent main class name.
		@param codebase the codebase for loading the agent classes
		(interpreted by the specified agent system).
		@param name the name of the new agent.
		@param place the name of place where the agent will be located.
		@param profile new agent profile. The agent type must be compatible
		with the agent system type of the target agent system.
		@param agent_data serializable object that will be passed as an argument
		to the afterBirth() agent method, unless the arguments parameter is not null.
		@param arguments array of arguments that will be passed as an argument
		to the afterBirth() agent method. Arguments must be Strings or object wrappers
		for Java primitive types. If this parameter is null, then the agent_data
		parameter is considered instead.
		@throws BadOperation agent creation failed.
		@see MobileObject#afterBirth(AgentSystem, AgentInfo, Object)
	*/
	public org.objectweb.mobilitools.smi.api.Name createAgent(
		Location location,
		String classname,
		String codebase,
		org.objectweb.mobilitools.smi.api.Name name,
		String place,
		AgentProfile profile,
		Serializable agent_data,
		Object argObj)
		throws BadOperation
	{
		org.objectweb.mobilitools.smi.api.Name result = null;
		MAFAgentSystem agency = resolveAgentSystem(location);
		if (agency != null)
		{
			byte[] argBytes;
			byte[] data;
			if (argObj == null)
			{
				argBytes = new byte[0];
				if (agent_data != null)
				{
					try
					{
						ByteArrayOutputStream dataBaos = new ByteArrayOutputStream();
						ObjectOutputStream objOos = new ObjectOutputStream(dataBaos);
						objOos.writeObject(agent_data);
						data = dataBaos.toByteArray();
					}
					catch (IOException e)
					{
						throw new BadOperation(
							BadOperation.SERIALIZATION,
							"Supplied data for agent creation cannot be serialized:\n" + e);
					}
				}
				else
				{
					data = new byte[0];
				}
			}
			else
			{
				data = new byte[0];
				try
				{
					ByteArrayOutputStream dataBaos = new ByteArrayOutputStream();
					ObjectOutputStream objOos = new ObjectOutputStream(dataBaos);
					objOos.writeObject(argObj);
					argBytes = dataBaos.toByteArray();
				}
				catch (IOException e)
				{
					throw new BadOperation(
						BadOperation.SERIALIZATION,
						"Supplied argument for agent creation cannot be serialized:\n" + e);
				}
			}
			try
			{
				result = new org.objectweb.mobilitools.smi.api.Name(
					agency.create_agent(
						name.getmafname(),
						profile,
						data,
						place,
						argBytes,
						new ClassName[] {
							new ClassName(
								classname,
								new byte[0])
						},
						codebase,
						null));
			}
			catch (SystemException e)
			{
				e.printStackTrace();
				throw new BadOperation(BadOperation.INFRASTRUCTURE, "CORBA SystemException while talking to agency " + location + ":\n" + e);
			}
			catch (MAFExtendedException e)
			{
				throw new BadOperation(
					BadOperation.OTHER,
					"Agent creation failed, probably for Java security or access modifiers reasons.");
			}
			catch (ClassUnknown e)
			{
				throw new BadOperation(
					BadOperation.CLASSFAULT,
					"Agent could not be created because of undefined class(es).");
			}
			catch (ArgumentInvalid e)
			{
				throw new BadOperation(
					BadOperation.REJECTED,
					"The agent refused to settle.");
			}
			catch (DeserializationFailed e)
			{
				throw new BadOperation(
					BadOperation.SERIALIZATION,
					"Provided data for agent creation could not be deserialized.");
			}
		}
		return result;
	}


	/**
		Moves an agent to the named agency in the specified region.
		@param from_region the name of the region the agent is currently residing in.
		@param agent the name of the agent.
		@param to_region the destination region.
		@param to_agency the name of the destination agency in the given destination region.
		@param place the destination place.
		@throws BadOperation agent move has failed.
	*/
	public void moveAgent(
		String from_region,
		org.objectweb.mobilitools.smi.api.Name agent,
		String to_region,
		org.objectweb.mobilitools.smi.api.Name to_agency,
		String place)
		throws BadOperation
	{
		moveAgent(
			findAgent(from_region, agent),
			agent,
			findAgency(to_region, to_agency),
			place);
	}


	/**
		Moves an agent to the specified location.
		@param from_location current agent location.
		@param agent the name of the agent.
		@param to_location the destination location.
		@param place the destination place.
		@throws BadOperation agent move has failed.
	*/
	public void moveAgent(Location from_location, org.objectweb.mobilitools.smi.api.Name agent, Location to_location, String place)
		throws BadOperation
	{
		MAFAgentSystemComplement agency = resolveAgentSystemComplement(from_location);
		try
		{
			agency.move_agent(agent.getmafname(), to_location.toString(), place);
		}
		catch (SystemException e)
		{
			e.printStackTrace();
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"CORBA SystemException while talking to agency " + from_location + ":\n" + e);
		}
		catch (ClassUnknown e)
		{
			throw new BadOperation(
				BadOperation.CLASSFAULT,
				"At least one class of the moving agent is not defined in destination agency.");
		}
		catch (DeserializationFailed e)
		{
			throw new BadOperation(
				BadOperation.SERIALIZATION,
				"Agent deserialization failed.");
		}
		catch (ArgumentInvalid e)
		{
			throw new BadOperation(
				BadOperation.DESTINATION,
				"Unknown destination agency " + to_location + " (bad location format).");
		}
		catch (AgentRefusedToMove e)
		{
			throw new BadOperation(
				BadOperation.REJECTED,
				"Agent refused to move.");
		}
		catch (AgentSystemNotFound e)
		{
			throw new BadOperation(
				BadOperation.DESTINATION,
				"Destination agency is not known by source agency.");
		}
		catch (AgentNotFound e)
		{
			throw new BadOperation(
				BadOperation.RETRY,
				"Agent is not present at its known host agency (it must have moved - retry).");
		}
		catch (MAFExtendedException e)
		{
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"CORBA communication or naming service problem in agent's host agency.");
		}
	}


	/**
		Resumes the activity of the specified agent in the specified region.
		@param region the name of the region where the agent is residing.
		@param agent the name wrapper of the agent MAF name.
		@throws BadOperation either the agent was already running, or its resumption failed.
	*/
	public void resumeAgent(String region, org.objectweb.mobilitools.smi.api.Name agent)
		throws BadOperation
	{
		resumeAgent(findAgent(region, agent), agent);
	}


	/**
		Resumes the activity of the specified agent residing at the specified location.
		@param location agent location.
		@param agent the name wrapper of the agent MAF name.
		@throws BadOperationeither the agent was already running, or its resumption failed.
	*/
	public void resumeAgent(Location location, org.objectweb.mobilitools.smi.api.Name agent)
		throws BadOperation
	{
		MAFAgentSystem agency;
		agency = resolveAgentSystem(location);
		try
		{
			agency.resume_agent(agent.getmafname());
		}
		catch (SystemException e)
		{
			e.printStackTrace();
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"CORBA SystemException while talking to agency " + location + ":\n" + e);
		}
		catch (ResumeFailed e)
		{
			throw new BadOperation(
				BadOperation.REJECTED,
				"The agent could resume its activity.");
		}
		catch (AgentIsRunning e)
		{
			throw new BadOperation(
				BadOperation.RUNNING,
				"The agent was already running.");
		}
		catch (AgentNotFound e)
		{
			throw new BadOperation(
				BadOperation.RETRY,
				"Agent is not present at its known host agency (it must have moved - retry).");
		}
	}


	/**
		Suspends the activity of the specified agent in the specified region.
		@param region the name of the region where the agent is residing.
		@param agent the name wrapper of the agent MAF name.
		@throws BadOperation either the agent was already suspended, or it could not be suspended.
	*/
	public void suspendAgent(String region, org.objectweb.mobilitools.smi.api.Name agent)
		throws BadOperation
	{
		suspendAgent(findAgent(region, agent), agent);
	}


	/**
		Suspends the activity of the specified agent residing at the specified location.
		@param location agent location.
		@param agent the name wrapper of the agent MAF name.
		@throws BadOperation either the agent was already suspended, or it could not be suspended.
	*/
	public void suspendAgent(Location location, org.objectweb.mobilitools.smi.api.Name agent)
		throws BadOperation
	{
		MAFAgentSystem agency;
		agency = resolveAgentSystem(location);
		try
		{
			agency.suspend_agent(agent.getmafname());
		}
		catch (SystemException e)
		{
			e.printStackTrace();
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"CORBA SystemException while talking to agency " + location + ":\n" + e);
		}
		catch (SuspendFailed e)
		{
			throw new BadOperation(
				BadOperation.REJECTED,
				"The agent could suspend its activity.");
		}
		catch (AgentIsSuspended e)
		{
			throw new BadOperation(
				BadOperation.SUSPENDED,
				"The agent was already suspended.");
		}
		catch (AgentNotFound e)
		{
			throw new BadOperation(
				BadOperation.RETRY,
				"Agent is not present at its known host agency (it must have moved - retry).");
		}
	}


	/**
		@return true if the specified agent in the specified region is currently active.
		@param region the name of the region where the agent is residing.
		@param agent wrapper of the agent MAF name.
		@throws BadOperation could not get agent activity status.
	*/
	public boolean isRunning(String region, org.objectweb.mobilitools.smi.api.Name agent)
		throws BadOperation
	{
		return isRunning(findAgent(region, agent), agent);
	}


	/**
		@return true if the specified agent in the specified region is currently active.
		@param location agent location.
		@param agent the name wrapper of the agent MAF name.
		@throws BadOperation could not get agent activity status.
	*/
	public boolean isRunning(Location location, org.objectweb.mobilitools.smi.api.Name agent)
		throws BadOperation
	{
		MAFAgentSystem agency;
		agency = resolveAgentSystem(location);
		try
		{
			return agency.get_agent_status(agent.getmafname()) == AgentStatus.CfMAFRunning;
		}
		catch (SystemException e)
		{
			e.printStackTrace();
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"CORBA SystemException while talking to agency " + location + ":" + e);
		}
		catch (AgentNotFound e)
		{
			throw new BadOperation(
				BadOperation.RETRY,
				"Agent is not present at its known host agency (it must have moved - retry).");
		}
	}


	/**
		Terminates the specified agent in the specified region.
		@param region the name of the region where the agent is residing.
		@param agent the name wrapper of the agent MAF name.
		@throws BadOperation
	*/
	public void terminateAgent(String region, org.objectweb.mobilitools.smi.api.Name agent)
		throws BadOperation
	{
		terminateAgent(findAgent(region, agent), agent);
	}


	/**
		Terminates the specified agent in the specified region.
		@param location agent location.
		@param agent the name wrapper of the agent MAF name.
		@throws BadOperation
	*/
	public void terminateAgent(Location location, org.objectweb.mobilitools.smi.api.Name agent)
		throws BadOperation
	{
		MAFAgentSystem agency = resolveAgentSystem(location);
		try
		{
			agency.terminate_agent(agent.getmafname());
		}
		catch (SystemException e)
		{
			e.printStackTrace();
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"CORBA SystemException while talking to agency " + location + ":\n" + e);
		}
		catch (AgentNotFound e)
		{
			throw new BadOperation(
				BadOperation.RETRY,
				"Agent " + agent + " is not present at its known host agency (it must have moved - retry).");
		}
		catch (TerminateFailed e)
		{
			throw new BadOperation(
				BadOperation.REJECTED,
				"Termination failed for agent " + agent + ".");
		}
	}


	/**
		Terminates the specified agency in the specified region.
		@param region the name of the region where the agency belongs to.
		@param agency wrapper of the agency MAF name.
		@throws BadOperation agency could not be terminated.
	*/
	public void terminateAgency(String region, org.objectweb.mobilitools.smi.api.Name agency)
		throws BadOperation
	{
		terminateAgency(findAgency(region, agency));
	}


	/**
		Terminates the specified agency in the specified region.
		@param location agency location.
		@throws BadOperation agency could not be terminated
	*/
	public void terminateAgency(Location location)
		throws BadOperation
	{
		MAFAgentSystem agency;
		agency = resolveAgentSystem(location);
		try
		{
			agency.terminate_agent_system();
		}
		catch (SystemException e)
		{
			e.printStackTrace();
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"CORBA SystemException while talking to agency " + location + ":\n" + e);
		}
		catch (TerminateFailed e)
		{
			throw new BadOperation(
				BadOperation.REJECTED,
				"Agency at " + location + " refused to terminate.");
		}
	}


	/**
		@return an array of agency locations where agents matching the specified properties
		reside, in the specified region.
		This array is zero-sized if no agent matches the given properties (or if there is no agent at all).
		@param region target region
		@param properties agent properties to match; may be null if all agents are wanted.
		@throws BadOperation could not get list of agent location
	*/
	public Location[] listAgentLocations(String region, Properties properties)
		throws BadOperation
	{
		Location[] result;
		MAFFinder finder = resolveFinder(region);
		try
		{
			AgentProfile profile = Misc.newDummyAgentProfile();
			profile.properties = Misc.property2any(properties, my_orb);
			String[] locations = finder.lookup_agent(
				Misc.DUMMYNAME,
				profile);
			result = new Location[locations.length];
			for (int i=0 ; i<locations.length ; ++i)
			{
				result[i] = new Location(locations[i]);
			}
		}
		catch (SystemException e)
		{
			e.printStackTrace();
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"CORBA SystemException while talking to region " + region + "'s MAFFinder:\n" + e);
		}
		catch (BadOperation e)
		{
			e.printStackTrace();
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"Unsupported location format found in MAFFinder (" + e.getMessage() + ")");
		}
		catch (EntryNotFound e)
		{
			result = new Location[0];
		}
		return result;
	}


	/**
		@return an array of names of agents matching the specified properties in the specified region.
		The returned array is zero-sized if no agent matching the given properties could be found.
		@param region target region.
		@param properties agent properties to match; may be null if all agents are wanted.
		@throws BadOperation could not get agents list.
	*/
	public org.objectweb.mobilitools.smi.api.Name[] listAgentNames(String region, Properties properties)
		throws BadOperation
	{
		org.objectweb.mobilitools.smi.api.Name[] result;
		MAFFinderComplement finder = resolveFinderComplement(region);
		try
		{
			AgentProfile profile = Misc.newDummyAgentProfile();
			profile.properties = Misc.property2any(properties, my_orb);
			org.omg.CfMAF.Name[] names = finder.lookup_agent_names(profile);
			result = new org.objectweb.mobilitools.smi.api.Name[names.length];
			for (int i=0 ; i<names.length ; ++i)
			{
				result[i] = new org.objectweb.mobilitools.smi.api.Name(names[i]);
			}
		}
		catch (SystemException e)
		{
			e.printStackTrace();
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"CORBA SystemException while talking to region " + region + "'s MAFFinder:" + e);
		}
		return result;
	}


	/**
		@return the names of all the agents at the specified agency location.
		The returned array is zero-sized if there is no agent at given location.
		@param location target agent system location
		@throws BadOperation could not get agents list.
	*/
	public org.objectweb.mobilitools.smi.api.Name[] listAgentsIn(Location location)
		throws BadOperation
	{
		org.objectweb.mobilitools.smi.api.Name[] result;
		MAFAgentSystem agentSystem = resolveAgentSystem(location);
		org.omg.CfMAF.Name[] names = null;
		try
		{
			names = agentSystem.list_all_agents();
			result = new org.objectweb.mobilitools.smi.api.Name[names.length];
			for (int i=0 ; i<names.length ; ++i)
			{
				result[i] = new org.objectweb.mobilitools.smi.api.Name(names[i]);
			}
		}
		catch (SystemException e)
		{
			e.printStackTrace();
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"CORBA SystemException while talking to agency " + location + ":\n" + e);
		}
		return result;
	}


	/**
		@return the names of the agents matching a given profile at the specified agency location.
		The returned array is zero-sized if there is no agent at given location.
		@param location target agent system location
		@param profile agent profile to match.
		@throws BadOperation could not get agents list
	*/
	public org.objectweb.mobilitools.smi.api.Name[] listAgentsIn(Location location, AgentProfile profile)
		throws BadOperation
	{
		org.objectweb.mobilitools.smi.api.Name[] result;
		MAFAgentSystemComplement agentSystem = resolveAgentSystemComplement(location);
		org.omg.CfMAF.Name[] names = null;
		try
		{
			names = agentSystem.list_agents(profile);
			result = new org.objectweb.mobilitools.smi.api.Name[names.length];
			for (int i=0 ; i<names.length ; ++i)
			{
				result[i] = new org.objectweb.mobilitools.smi.api.Name(names[i]);
			}
		}
		catch (SystemException e)
		{
			e.printStackTrace();
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"CORBA SystemException while talking to agency " + location + ":\n" + e);
		}
		return result;
	}


	/**
		@return the locations of all the agent systems matching the given properties
		in the specified region.
		The returned array is zero-sized if no agent system matches the properties (or if there is no agent system at all).
		@param region target region.
		@param properties properties that selected agencies must match;
		may be null if all agencies are wanted.
		@throws BadOperation could not get agencies list
	*/
	public Location[] listAgencies(String region, Properties properties)
		throws BadOperation
	{
		Location[] result;
		MAFFinder finder = resolveFinder(region);
		try
		{
			AgentSystemInfo info = Misc.newDummyAgentSystemInfo();
			info.properties = Misc.property2any(properties, my_orb);
			String[] locations = finder.lookup_agent_system(
				Misc.DUMMYNAME,
				info);
			result = new Location[locations.length];
			for (int i=0 ; i<locations.length ; ++i)
			{
				result[i] = new Location(locations[i]);
			}
		}
		catch (SystemException e)
		{
			e.printStackTrace();
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"CORBA SystemException while talking to region " + region + "'s MAFFinder:\n" + e);
		}
		catch (EntryNotFound e)
		{
			result = new Location[0];
		}
		catch (BadOperation e)
		{
			e.printStackTrace();
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"Unsupported location format found in MAFFinder (" + e.getMessage() + ").");
		}
		return result;
	}


	/**
		@return the location of the named agent residing at the specified region.
		@param region target region.
		@param agent name wrapper of agent MAF name.
		@throws BadOperation the agent could not be found.
	*/
	public Location findAgent(String region, org.objectweb.mobilitools.smi.api.Name agent)
		throws BadOperation
	{
		MAFFinder finder = resolveFinder(region);
		try
		{
			return new Location(
				finder.lookup_agent(
					agent.getmafname(),
					Misc.DUMMYAGENTPROFILE)[0]);
		}
		catch (SystemException e)
		{
			e.printStackTrace();
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"CORBA SystemException while talking to region " + region + "'s MAFFinder:\n" + e);
		}
		catch (EntryNotFound e)
		{
			throw new BadOperation(
				BadOperation.UNKNOWNAGENT,
				"Unknown agent " + agent + " in region " + region + ".");
		}
		catch (BadOperation e)
		{
			e.printStackTrace();
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"Unsupported location format found in MAFFinder (" + e.getMessage() + ").");
		}
	}


	/**
		@return the location of the named agent system belonging to the specified region.
		@param region target region.
		@param agent name wrapper of agent system MAF name.
		@throws BadOperation the agency could not be found.
	*/
	public Location findAgency(String region, org.objectweb.mobilitools.smi.api.Name agency)
		throws BadOperation
	{
		MAFFinder finder = resolveFinder(region);
		try
		{
			return new Location(
				finder.lookup_agent_system(
					agency.getmafname(),
					Misc.DUMMYAGENTSYSTEMINFO)[0]);
		}
		catch (SystemException e)
		{
			e.printStackTrace();
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"CORBA SystemException while talking to region " + region + "'s MAFFinder:\n" + e);
		}
		catch (EntryNotFound e)
		{
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"Unknown agency " + agency + " in region " + region + ".");
		}
		catch (BadOperation e)
		{
			e.printStackTrace();
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"Unsupported location format found in MAFFinder (" + e.getMessage() + ").");
		}
	}


	/**
		@return the properties of the named agent residing in the specified region.
		@param region target region.
		@param name wrapper of agent MAF name.
		@exception BadOperation agent properties could not be obtained.
	*/
	public Properties getAgentProperties(String region, org.objectweb.mobilitools.smi.api.Name name)
		throws BadOperation
	{
		MAFFinderComplement finder = null;
		try
		{
			finder = MAFFinderComplementHelper.narrow(resolveFinder(region));
		}
		catch (SystemException e)
		{
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"The MAFFinder for region " + region + " does not implement MAFFinderComplement.");
		}
		try
		{
			return Misc.any2property(finder.get_agent_profile(name.getmafname()).properties);
		}
		catch (SystemException e)
		{
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"CORBA SystemException while talking to region " + region + "'s MAFFinder:\n" + e);
		}
		catch (AgentNotFound e)
		{
			throw new BadOperation(
				BadOperation.RETRY,
				"Agent is not present at its known host agency (it must have moved - retry).");
		}
	}


	/**
		@return the CORBA object reference to the MAFAgentSystem implementation
		at the specified location.
		@param location agent system location.
		@throws BadOperation the location did not resolve to an agent system.
	*/
	MAFAgentSystem resolveAgentSystem(Location location)
		throws BadOperation
	{
		MAFAgentSystem result = null;
		try
		{
			result = MAFAgentSystemHelper.narrow(my_ns.resolve(location.getCosNamingName()));
		}
		catch (NameServiceException e)
		{
			if (e.exception instanceof SystemException)
			{
				e.printStackTrace();
				throw new BadOperation(
					BadOperation.INFRASTRUCTURE,
					"CORBA SystemException while talking to naming service:\n" + e);
			}
			else
			{
				throw new BadOperation(
					BadOperation.UNKNOWNAGENCY,
					"Unknown agency " + location);
			}
		}
		catch (SystemException e)
		{
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"CORBA SystemException while trying to reach agency " + location + ":\n" + e);
		}
		if (result == null)
		{
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"Could not narrow object to a MAFAgentSystem implementation.");
		}
		return result;
	}


	/**
		@return the CORBA object reference to the MAFAgentSystemComplement implementation
		at the specified location.
		@param location agent system location.
		@throws BadOperation the location did not resolve to an extended agent system.
	*/
	MAFAgentSystemComplement resolveAgentSystemComplement(Location location)
		throws BadOperation
	{
		MAFAgentSystemComplement result = null;
		try
		{
			result = MAFAgentSystemComplementHelper.narrow(resolveAgentSystem(location));
		}
		catch (SystemException e)
		{
			e.printStackTrace();
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"Agent's host agency is unreachable or does not implement MAFAgentSystemComplement.");
		}
		if (result == null)
		{
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"Could not narrow object to a MAFAgentSystemComplement implementation.");
		}
		return result;
	}


	/**
		@return the CORBA object reference to the MAFFinder implementation
		for the specified region.
		@param region the name of the region the requested MAFFinder is bound to.
		@throws BadOperation could not find a MAFFinder implementation for given region.
	*/
	MAFFinder resolveFinder(String region)
		throws BadOperation
	{
		MAFFinder result = null;
		String finderName =
			System.getProperty(Constants.regionPrefixProp, Constants.regionPrefixDefault) +
			"/" + region + "/" +
			System.getProperty(Constants.finderNameProp, Constants.finderNameDefault);
		try
		{
			result = MAFFinderHelper.narrow(my_ns.resolve(finderName));
		}
		catch (NameServiceException e)
		{
			if (e.exception instanceof SystemException)
			{
				e.printStackTrace();
				throw new BadOperation(
					BadOperation.INFRASTRUCTURE,
					"CORBA SystemException while talking to naming service:\n" + e);
			}
			else
			{
				throw new BadOperation(
					BadOperation.INFRASTRUCTURE,
					"It seems that region " + region + " does not have a MAF Finder.");
			}
		}
		catch (SystemException e)
		{
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"CORBA SystemException while trying to reach region " + region + "'s MAF Finder:\n" + e);
		}
		if (result == null)
		{
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"Can not resolve object to a MAFFinder.");
		}
		return result;
	}


	/**
		@return the CORBA object reference to the MAFFinderComplement implementation
		for the specified region.
		@param region the name of the region the requested MAFFinder is bound to.
		@throws BadOperation could not find a MAFFinderComplement implementation for given region.
	*/
	MAFFinderComplement resolveFinderComplement(String region)
		throws BadOperation
	{
		MAFFinderComplement result = null;
		try
		{
			result = MAFFinderComplementHelper.narrow(resolveFinder(region));
		}
		catch (SystemException e)
		{
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"The MAFFinder for region " + region + " does not implement MAFFinderComplement.");
		}
		if (result == null)
		{
			throw new BadOperation(
				BadOperation.INFRASTRUCTURE,
				"Can not nerrow object to a MAFFinderComplement implementation.");
		}
		return result;
	}
}
