/*
* MobiliTools: an implementation of the Object Management Group's
* Mobile Agent Facility specification.
* Copyright (C) 2003 France Telecom R&D
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
* MobiliTools $Name:  $
*
* Contact: mobilitools-smi@lists.debian-sf.objectweb.org
*
* Authors: Bruno Dillenseger
*/


package org.objectweb.mobilitools.smi.lib;


import java.util.Hashtable;


/**
 * MobiliTools $Name:  $, $Id: ClassCache.java,v 1.1.1.1 2003/03/28 14:48:06 dillense Exp $
 * <P>
 * Cache for classes, bytecode and any resource associated to a given codebase.
 * A cache singleton is managed for each codebase value.
 * A counter is maintained for each so that the cache is cleared when its
 * codebase is not used by any agent any more.
*/
public class ClassCache extends Hashtable
{
	static protected Hashtable caches = new Hashtable();
	protected volatile long refCount=0;


	/**
		Ensure that a cache exists for the given codebase - creating a new cache instance
		if necessary - and returns its reference. In other words, there is a cache
		singleton for each given codebase value.
		@param codebase the codebase associated to the cache
		@return the cache associated to the given codebase
	*/
	static public synchronized ClassCache getCache(String codebase)
	{
		ClassCache cache = (ClassCache)caches.get(codebase);
		if (cache == null)
		{
			cache = new ClassCache();
			caches.put(codebase, cache);
		}
		return cache;
	}


	/**
		Increments the reference counter of the cache associated to the given codebase.
		Does nothing if no cache exists for the given codebase.
		@param codebase the codebase of the target cache.
	*/
	static public synchronized void inc(String codebase)
	{
		ClassCache cache = getCache(codebase);
		synchronized (cache)
		{
			if (cache != null)
			{
				++cache.refCount;
			}
		}
	}


	/**
		Decrements the reference counter of the cache associated to the given codebase.
		If the reference counter's new value is zero, then the cache is cleared.
		Does nothing if no cache exists for the given codebase.
		@param codebase the codebase of the target cache.
	*/
	static public synchronized void dec(String codebase)
	{
		ClassCache cache = getCache(codebase);
		synchronized (cache)
		{
			--cache.refCount;
			if (cache.refCount == 0)
			{
				cache.clear();
				caches.remove(codebase);
			}
			else if (cache.refCount < 0)
			{
				throw new Error("cache.refCount negatif pour codebase " + codebase);
			}
		}
	}


	protected ClassCache()
	{
		super();
	}
}
