/*
* MobiliTools: an implementation of the Object Management Group's
* Mobile Agent Facility specification.
* Copyright (C) 2003 France Telecom R&D
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
* MobiliTools $Name:  $
*
* Contact: mobilitools-smi@lists.debian-sf.objectweb.org
*
* Authors: Bruno Dillenseger
*/


package org.objectweb.mobilitools.smi.lib;


import org.objectweb.mobilitools.smi.lib.bytearray.ByteArrayURLStreamHandler;
import org.omg.CfMAF.*;
import org.omg.CORBA.*;
import java.net.*;
import java.io.*;
import java.util.*;
import java.util.jar.*;


/**
 * MobiliTools $Name:  $, $Id: SMIClassLoader.java,v 1.1.1.1 2003/03/28 14:48:06 dillense Exp $
 * <P>
 * MAF-based class loader for SMI. Makes it possible to define classes by
 * loading their byte code from a local filesystem, a URL, or an implementation
 * of MAF's MAFAgentSystem interface. It also handles access to resources in
 * the same manner. A dedicated classloader is created for each new mobile
 * object (created or incoming). However, byte-code, classes and resources are
 * kept in a cache and reused for each given codebase. This way, mobile objects
 * may invoke each other through and only through a class, interface or object
 * with the same codebase, or loaded by a common parent class loader.
 * <P>
 * The class/resource loading policy is the following:
 * <OL>
 *   <LI>try to get the missing class/resource from the parent classloader
 *   <LI>try to retrieve the missing class/resource from the local cache
 *   <LI>ask the class provider (SMI agency), which, in turn, looks for the
 *   class byte-code or resource in its cache
 *   <LI>locally interpret the codebase
 * </OL>
 * A cache is flushed at garbage collection time if and only if no more agent
 * with the corresponding codebase is present in the agency.
 * <P>
 * Notes:
 * <UL>
 *   <LI>SMIClassLoader typically makes it possible to move agents even if
 *   their original class source (HTTP server, local file or whatever) is not
 *   reachable any more.
 *   <LI>if you wish to use a simple URL class loading behaviour, you may
 *   look at URLClassLoaderAdapter class
 *   <LI>if mobile objects use RMI, RMIClassLoaders should be used (see
 *   RMIClassLoaderAdapter class)
 *   <LI>to use custom classloaders, declare the corresponding factory in the
 *   appropriate system property. This classloader factory class must
 *   implement a static public method getClassLoader() returning the custom
 *   class loaders.
 * </UL>
 * @see RMIClassLoaderAdapter
 * @see URLClassLoaderAdapter
 * @see #getClassLoader(ClassLoader, String, AgentProfile, MAFAgentSystem)
 * @see org.objectweb.mobilitools.smi.api.Constants
*/
public class SMIClassLoader extends ClassLoader
{
	String my_codebase;
	AgentProfile my_profile;
	MAFAgentSystem my_provider;
	volatile ClassCache my_cache;


	/**
	 * Creates a new SMI class loader.
	 * @param the parent class loader (from which SMI class loader tries to get
	 * classes before fetching them by itself)
	 * @param codebase the codebase for the classes; may be a single pathname or
	 * URL designating a ZIP/JAR file or directory holding the classes
	 * @param profile the agent profile (based on SMI's profile, see MAF spec.)
	 * @param provider the MAFAgentSystem implementation (SMI agency) providing
	 * the classes.
	 * @return each call returns a new classloader
	 */
	static public ClassLoader getClassLoader(
		ClassLoader parent,
		String codebase,
		AgentProfile profile,
		MAFAgentSystem provider)
	{
		return new SMIClassLoader(parent, codebase, profile, provider);
	}


	/**
		Reads or retrieves from a cache the byte code for a given class name, or
		a resource content (accordingly to current class loader's code base).
		There is one cache per represented codebase. Caches are filled by the findClass()
		instance method, when a class is downloaded from another
		agency by the fetch_class() operation.
		@param name the class full name (i.e. including the package name if any) or resource name
		@return the byte code or the resource content designated by the name, or
		null if the byte code or resource could not be found.
	*/
	public byte[] getBytes(String name)
	{
		synchronized (my_cache)
		{
			byte[] result = null;
			// try to get the class code from the cache (in case it has already been downloaded for current codebase)
			result = (byte[])my_cache.get(name);
			if (result == null && my_provider != null)
			{
				// invoke a MAFAgentSystem to get the bytes
				try
				{
					ClassName[] className = { new ClassName(name, new byte[0]) };
					result = my_provider.fetch_class(className, my_codebase, my_profile)[0];
				}
				catch (Exception ex)
				{
					System.err.println("warning: could not get bytes from class provider (" + ex + ") => trying codebase interpretation");
				}
			}
			if (result == null)
			{
				// get the class code from a JAR or class file, designated by a local file path, or a URL
				try
				{
					// is the codebase a URL?
					URL url = null;
					try
					{
						url = new URL(my_codebase);
					}
					catch (MalformedURLException e)
					{
					}
					InputStream is = null;
					boolean is_jar = false;
					JarInputStream jis = null;
					JarEntry je = null;
					if (url == null)
					// the codebase is local path
					{
						if (my_codebase.endsWith("/"))
						// the codebase is a local directory (not a JAR file)
						{
							is = new FileInputStream(my_codebase + name);
						}
						else
						// the codebase may be a JAR file
						{
							try
							{
								is = new FileInputStream(my_codebase);
								jis = new JarInputStream(is);
								je = jis.getNextJarEntry();
							}
							catch (FileNotFoundException e)
							{
							}
							if (je == null)
							// the codebase is not a JAR file (or an empty JAR we ignore)
							{
								is = new FileInputStream(my_codebase + "/" + name);
							}
							else
							// the codebase is a JAR file
							{
								is_jar = true;
							}
						}
					}
					else
					// the codebase is a URL
					{
						if (my_codebase.endsWith("/"))
						// the codebase URL designates a directory
						{
							url = new URL(url, name);
							is = url.openConnection().getInputStream();
						}
						else
						// the codebase URL may designate a JAR file
						{
							try
							{
								is = url.openConnection().getInputStream();
								jis = new JarInputStream(is);
								je = jis.getNextJarEntry();
							}
							catch (IOException e)
							{
							}
							if (je == null)
							// the codebase URL is not a JAR file (or an empty JAR we ignore)
							{
								url = new URL(my_codebase + "/" + name);
								is = url.openConnection().getInputStream();
							}
							else
							// the codebase URL is a JAR file
							{
								is_jar = true;
							}
						}
					}
					// get an input stream from the byte code file or URL
					BufferedInputStream bis = null;
					if (is_jar)
					// get the right entry from the JAR
					{
						while (je != null && !je.getName().equals(name))
						{
							je = jis.getNextJarEntry();
						}
						if (je != null)
						{
							bis = new BufferedInputStream(jis);
						}
					}
					else
					{
						bis = new BufferedInputStream(is);
					}
					// load byte code and tranfer it to a byte array
					ByteArrayOutputStream baos = new ByteArrayOutputStream();
					int n;
					while ((n = bis.read()) != -1)
					{
						baos.write(n);
					}
					result = baos.toByteArray();
				}
				catch (Exception e)
				{
				}
			}
			if (result != null)
			{
				my_cache.put(name, result);
			}
			return result;
		}
	}


	/**
		Creates a new SMI class loader, able to load classes from the local filesystem, URLs,
		and a MAFAgentSystem inplementation.
		@param parent a classloader able to load classes from the classpath.
		If null, the system class loader will be used for that purpose.
		@param codebase the SMI class loader's codebase (single filesystem
		pathname or URL designating a ZIP/JAR file or a directory)
		@param profile the agent profile (see MAF specifications)
		@param provider the remote MAFAgentSystem that may provide the missing classes.
		Typically, the provider is the agency where a incoming agent is migrating from.
	*/
	SMIClassLoader(ClassLoader parent, String codebase, AgentProfile profile, MAFAgentSystem provider)
	{
		super(parent);
		my_codebase = codebase;
		my_profile = profile;
		my_provider = provider;
		my_cache = ClassCache.getCache(codebase);
		ClassCache.inc(codebase);
	}


	/**
		Gets the byte code for, and defines the requested class, either through
		local or remote bytecode load.
		Byte-code remotely loaded through the fetch_class() operation of a MAFAgentSystem
		implementation is cached. This
		cache is necessary for the next migration of the incoming agent, in order to
		provide the next hosting agency with the class bytecode. Classes in the cache are
		identified by their name and codebase.
		@param name the name of the class to define (see Java2 API).
		@return the corresponding Class object.
		@exception ClassNotFoundException if the class could not be loaded and defined.
	*/
	protected Class findClass(String name)
		throws ClassNotFoundException
	{
		synchronized (my_cache)
		{
			Class result = null;
			String pathname = name.replace('.', '/') + ".class";
			// first, see if the class can be found using the classpath
			try
			{
				ClassLoader parent = getParent();
				if (parent == null)
				{
					parent = getSystemClassLoader();
				}
				result = parent.loadClass(name);
			}
			catch (ClassNotFoundException e)
			{
				// second, see if the class has already been defined for current codebase
				result = (Class)my_cache.get(name);
				if (result == null)
				{
					// then, try to use the codebase and the provider
					byte[] code = getBytes(pathname);
					if (code != null)
					{
						result = defineClass(name, code, 0, code.length);
						my_cache.put(name, result);
					}
					else
					{
						throw new ClassNotFoundException(name);
					}
				}
			}
			return result;
		}
	}


	/**
	 * Gets the content of the named resource and keeps it in the cache
	 * associated to current SMI class loader's codebase.
	 * @param name the resource's name
	 * @return a URL with a specific scheme to access the resource content, or
	 * null if the resource could not be found/loaded.
	 */
	public URL findResource(String name)
	{
		try
		{
			return new URL(null, "bytearray:" + my_codebase + "#" + name, new ByteArrayURLStreamHandler(this));
		}
		catch (Exception ex)
		{
			ex.printStackTrace();
			return null;
		}
	}


	/**
	 * Since a codebase for SMI class loaders hold a single value, this method
	 * is similar to findResource().
	 * @param name the resource's name
	 * @return an empty or single-value enumeration of URLs for the named
	 * resource
	 * @see #findResource(String)
	 */
	public Enumeration findResources(String name)
	{
		Vector result = new Vector(1);
		URL url = findResource(name);
		if (url != null)
		{
			result.add(url);
		}
		return result.elements();
	}


	public String toString()
	{
		return "SMIClassLoader for codebase " + my_codebase;
	}


	public void finalize()
	{
		ClassCache.dec(my_codebase);
	}
}
