/*
* MobiliTools: an implementation of the Object Management Group's
* Mobile Agent Facility specification.
* Copyright (C) 2003 France Telecom R&D
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
* MobiliTools $Name:  $
*
* Contact: mobilitools-smi@lists.debian-sf.objectweb.org
*
* Authors: Bruno Dillenseger
*/


package org.objectweb.mobilitools.util.corba;


import java.io.*;
import java.net.URL;
import java.util.Properties;
import org.omg.CORBA.ORB;
import mparser.*;


/**
 * MobiliTools $Name:  $, $Id: NSbinder.java,v 1.1.1.1 2003/03/28 14:48:06 dillense Exp $
 * <P>
 * Tool for making automatic bindings in CORBA naming service.
 * The input can be a property file, an XML file, or Java properties.
 * Binding names must be given according to "id!kind/.../id!kind" format.
 * A binding object may be given as a stringified IOR, as a file containing
 * a stringified IOR, or as a URL containing a stringified IOR.
 */
public class NSbinder
{
	NameService ns;
	ORB orb;


	/**
		Parses a property file or an xml file describing a set
		of CORBA name service bindings to do.
		@param args shall contain ORB-specific switches first, and
		then either <code>-xml xml_file</code> or <code>-prop property_file</code>
		switch.
		@see #bindFromXMLFile(String)
		@see #bindFromPropFile(String)
	*/
	static public void main(String[] args)
	{
		try
		{
			NSbinder binder = new NSbinder(ORB.init(args, System.getProperties()));
			if (args[args.length-2].equalsIgnoreCase("-prop"))
			{
				binder.bindFromPropFile(args[args.length-1]);
			}
			else if(args[args.length-2].equalsIgnoreCase("-xml"))
			{
				binder.bindFromXMLFile(args[args.length-1]);
			}
			else
			{
				System.out.println("usage: [ORB-specific args...] [-prop file] [-xml file]");
			}
		}
		catch (Throwable ex)
		{
			System.err.println(ex.toString());
		}
	}


	/**
		Creates a new name service binder utility, using default/singleton ORB object.
		@throws NameServiceException if no valid initial reference to a naming service
		could be obtained.
	*/
	public NSbinder()
		throws NameServiceException
	{
		this(ORB.init());
	}


	/**
		Creates a new name service binder utility, using the provided ORB object.
		@param orb the ORB object to use.
		@throws NameServiceException if no valid initial reference to a naming service
		could be obtained.
	*/
	public NSbinder(ORB orb)
		throws NameServiceException
	{
		this.orb = orb;
		ns = new NameService(orb);
	}


	/**
		Parses the specified XML file in order to make the specified bindings.
		This file should contain only bind tags, with the following attributes:
		<UL>
			<LI><code>&lt;bind name="some_new_context" /&gt;</code> <I>rebinds name to a new naming context</I>
			<LI><code>&lt;bind name="some_object" ior="IOR:...." /&gt;</code> <I>rebinds name to the specified IOR</I>
			<LI><code>&lt;bind name="some_object" file="IOR_file" /&gt;</code> <I>rebinds name to IOR contained in the specified file</I>
			<LI><code>&lt;bind name="some_object" url="IOR_URL" /&gt;</code> <I>rebinds name to IOR contained at the specified URL</I>
		</UL>
		@param file the name of the XML file to be parsed.
		@throws Exception a variety of exceptions if a parsing error occurs,
		because of an invalid file format (check the XML content)
	*/
	public void bindFromXMLFile(String file)
		throws Exception
	{
		WMparser parser = new WMparser(file);
		ItemParsed item;
		Properties props = new Properties();
		int bindCount = 0;

		while (parser.hasMoreItems())
		{
			item = parser.getNextItem();
			System.out.println(item.getName());
			if ((item.getType() == ItemParsed.ITEM_TYPE_ELEMT) && item.getName().equalsIgnoreCase("bind"))
			{
				++bindCount;
			}
			else if (item.getType() == ItemParsed.ITEM_TYPE_ATTRB)
			{
				props.setProperty(
					item.getName().substring("bind.".length()) + "." + String.valueOf(bindCount),
					item.getValue());
			}
			else if (item.getType() == ItemParsed.ITEM_TYPE_TEXTE)
			{
				System.out.println(item.getValue());
			}
			else
			{
				System.out.println("XML deployment warning: ignoring \"" + item.getName() + "\"");
			}
		}
		props.list(System.out);
		bind(props);
	}


	/**
		Parses the specified property file in order to make the specified bindings.
		@param file the name of the property file to be parsed.
		@throws IOException if the specified file could not be read.
		@see #bind(Properties)
	*/
	public void bindFromPropFile(String file)
		throws IOException
	{
		Properties props = new Properties();
		props.load(new FileInputStream(new File(file)));
		bind(props);
	}


	/**
		Performs CORBA name service bindings specified by the provided properties.
		Properties are:
		<UL>
			<LI>name.<I>n</I> for <I>n</I>th binding name
			<LI>url.<I>n</I> for URL of <I>n</I>th binding IOR
			<LI>ior.<I>n</I> for <I>n</I>th stringified binding IOR
			<LI>file.<I>n</I> for file containing <I>n</I>th binding IOR
		</UL>
		Where <I>n</I> increments from 1 and stops as soon as property name.<I>n</I> is undefined.
		If no URL, file or IOR is specified for a name, then the name is bound to a new naming context.
	*/
	public void bind(Properties props)
	{
		String name, url, ior, file;
		String suffix;

		for (int i=1 ; (name = props.getProperty("name." + (suffix = String.valueOf(i)))) != null ; ++i)
		{
			url = props.getProperty("url." + suffix);
			ior = props.getProperty("ior." + suffix);
			file = props.getProperty("file." + suffix);
			try
			{
				if (url != null)
				{
					System.out.println("binding " + name + " to IOR from URL " + url);
					ns.rebind(
						name,
						orb.string_to_object((new BufferedReader(new InputStreamReader((new URL(url)).openStream()))).readLine()));
				}
				else if (file != null)
				{
					System.out.println("binding " + name + " to IOR from file " + file);
					ns.rebind(
						name,
						orb.string_to_object((new BufferedReader(new InputStreamReader((new URL("file:" + file)).openStream()))).readLine()));
				}
				else if (ior != null)
				{
					System.out.println("binding " + name + " to IOR " + ior);
					ns.rebind(name, orb.string_to_object(ior));
				}
				else
				{
					System.out.println("binding " + name + " to new context");
					ns.makePath(name);
				}
			}
			catch (Exception e)
			{
				System.err.println(e.toString());
			}
		}
	}
}
