/*
* MobiliTools: an implementation of the Object Management Group's
* Mobile Agent Facility specification.
* Copyright (C) 2003 France Telecom R&D
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
* MobiliTools $Name:  $
*
* Contact: mobilitools-smi@lists.debian-sf.objectweb.org
*
* Authors: Bruno Dillenseger
*/


package org.objectweb.mobilitools.util.corba;


import java.awt.*;
import org.omg.CORBA.*;
import java.awt.event.*;
import org.objectweb.mobilitools.util.gui.AlertWindow;

/**
 * MobiliTools $Name:  $, $Id: NSbrowser.java,v 1.1.1.1 2003/03/28 14:48:06 dillense Exp $
 * <P>
 * Browser for CORBA name service.
 * This GUI is based on the NameService wrapper (@see NameService).
 * It can be used:
 * <UL>
 *   <LI>either directly as a standalone programme, to browse the default context of the default CORBA name service;
 *   <LI>or from any programme by providing a NameService object to the constructor, to browse any name service from any context.
 * </UL>
 * Buttons:
 * <UL>
 *   <LI>"up", to go to the parent context;
 *   <LI>"down", to list the selected context (a double-click does the same);
 *   <LI>"update", to re-list the content of current context;
 *   <LI>"view", to see the string'fied IOR of the selected object or context;
 *   <LI>"bind", to bind an object or a context in current context. If a new context
 *   is created, then a name (id and kind) is sufficient. Otherwise, one must provide
 *   also the string'fied IOR of the object or context to bind.
 *   <LI>"unbind", to unbind the selected object or context. Note that removing a context
 *   doesn't unbind its children objects, and does not destroy the context.
 *   <LI>"destroy", to destroy the selected context, and destroy its children contexts recursively,
 *   while removing any binding in those contexts.
 * </UL>
 * @see NameService
*/
public class NSbrowser extends Frame implements ActionListener, ItemListener
{
	Button upBtn, downBtn, updateBtn, viewBtn, bindBtn, unbindBtn, destroyBtn;
	List bindingsLst;
	NameService my_ns;
	ORB my_orb;
	NSbinding[] bindings;
	String path;
	static boolean standalone = false;


	/**
		Opens a CORBA naming service GUI for the default context of the default name service
		(i.e. from the initial reference).
	*/
	static public void main(String[] args)
	{
		try
		{
			ORB orb = ORB.init(args, null);
			standalone = true;
			new NSbrowser(orb, new NameService(orb));
		}
		catch (NameServiceException e)
		{
			System.err.println(e.toString());
			e.printStackTrace();
		}
	}


	/**
		Opens a GUI for the specified NameService object.
		@param the_orb is a reference to the ORB object (see CORBA's ORB_init())
		@param the_ns the NameService object
		@see NameService
	*/
	NSbrowser(ORB the_orb, NameService the_ns)
	{
		my_orb = the_orb;
		my_ns = the_ns;
		path = "/";
		setLayout(new BorderLayout());
		Panel navigationPnl = new Panel();
		navigationPnl.setLayout(new FlowLayout());
		navigationPnl.add(upBtn = new Button("up"));
		upBtn.addActionListener(this);
		navigationPnl.add(downBtn = new Button("down"));
		downBtn.addActionListener(this);
		navigationPnl.add(updateBtn = new Button("update"));
		updateBtn.addActionListener(this);
		add("North", navigationPnl);
		add("Center", bindingsLst = new List());
		bindingsLst.addActionListener(this);
		bindingsLst.addItemListener(this);
		Panel editionPnl = new Panel();
		editionPnl.setLayout(new FlowLayout());
		editionPnl.add(viewBtn = new Button("view"));
		viewBtn.addActionListener(this);
		editionPnl.add(bindBtn = new Button("bind"));
		bindBtn.addActionListener(this);
		editionPnl.add(unbindBtn = new Button("unbind"));
		unbindBtn.addActionListener(this);
		editionPnl.add(destroyBtn = new Button("destroy"));
		destroyBtn.addActionListener(this);
		add("South", editionPnl);
		updateBindings();
		addWindowListener(new OnWindowClosing());
		pack();
		show();
	}


	void updateBindings()
	{
		try
		{
			bindings = my_ns.list(path);
			if (bindingsLst.getItemCount() > 0)
			{
				bindingsLst.removeAll();
			}
			for (int i=0 ; i<bindings.length ; ++i)
			{
				if (bindings[i].type == NSbinding.OBJECT)
				{
					bindingsLst.add(bindings[i].name + "!" + bindings[i].kind);
				}
				else
				{
					bindingsLst.add(bindings[i].name + "!" + bindings[i].kind + "/");
				}
			}
			setTitle(path);
			viewBtn.setEnabled(false);
			unbindBtn.setEnabled(false);
			destroyBtn.setEnabled(false);
			downBtn.setEnabled(false);
			if (path.equals("/"))
			{
				upBtn.setEnabled(false);
			}
			else
			{
				upBtn.setEnabled(true);
			}
		}
		catch (NameServiceException e)
		{
			new AlertWindow(this, "Name Service Exception", e.toString(), "OK");
			e.printStackTrace();
		}
	}


	////////////////////////////////////////////////
	// Implementation of interface ActionListener //
	////////////////////////////////////////////////


	public void actionPerformed(ActionEvent evt)
	{
		if (evt.getSource() == upBtn)
		{
			try
			{
				path = path.substring(0, 1 + path.substring(0, path.length()-1).lastIndexOf("/"));
			}
			catch (StringIndexOutOfBoundsException e)
			{
				path = "/";
			}
			updateBindings();
		}
		else if (evt.getSource() == downBtn || evt.getSource() == bindingsLst)
		{
			if (bindings[bindingsLst.getSelectedIndex()].type == NSbinding.CONTEXT)
			{
				path = path + bindingsLst.getSelectedItem();
				updateBindings();
			}
		}
		else if (evt.getSource() == updateBtn)
		{
			updateBindings();
		}
		else if (evt.getSource() == viewBtn)
		{
			new AlertWindow(
				this,
				path + bindingsLst.getSelectedItem(),
				my_orb.object_to_string(bindings[bindingsLst.getSelectedIndex()].object),
				"OK");
		}
		else if (evt.getSource() == bindBtn)
		{
			Dialog bindDlg = new Dialog(this, "Add a binding in " + path, true);
			TextArea refTxt;
			TextField idFld, kindFld;
			Checkbox contextChk;
			Button bindBtn, cancelBtn;
			Panel namePnl, buttonPnl;

			// GUI layout
			bindDlg.setLayout(new BorderLayout());
			namePnl = new Panel();
			namePnl.setLayout(new GridLayout(1,5));
			namePnl.add(new Label("id:"));
			namePnl.add(idFld = new TextField());
			namePnl.add(new Label("kind:"));
			namePnl.add(kindFld = new TextField());
			namePnl.add(contextChk = new Checkbox("new context", true));
			bindDlg.add("North", namePnl);
			bindDlg.add("Center", refTxt = new TextArea());
			refTxt.setEnabled(false);
			buttonPnl = new Panel();
			buttonPnl.setLayout(new FlowLayout());
			buttonPnl.add(bindBtn = new Button("bind"));
			buttonPnl.add(cancelBtn = new Button("cancel"));
			bindDlg.add("South", buttonPnl);

			// GUI listeners setup
			contextChk.addItemListener(new ContextChkMgr(refTxt));
			bindDlg.addWindowListener(new BindDlgMgr(bindDlg));
			bindBtn.addActionListener(new BindBtnMgr(bindDlg, idFld, kindFld, contextChk, refTxt));
			cancelBtn.addActionListener(new CancelBtnMgr(bindDlg));

			bindDlg.pack();
			bindDlg.show();
		}
		else if (evt.getSource() == unbindBtn)
		{
			try
			{
				my_ns.unbind(path + bindingsLst.getSelectedItem());
			}
			catch (NameServiceException e)
			{
				new AlertWindow(this, "Name Service Exception", e.toString(), "OK");
				e.printStackTrace();
			}
			updateBindings();
		}
		else if (evt.getSource() == destroyBtn)
		{
			try
			{
				my_ns.deepUnbind(path + bindingsLst.getSelectedItem());
			}
			catch (NameServiceException e)
			{
				new AlertWindow(this, "Name Service Exception", e.toString(), "OK");
				e.printStackTrace();
			}
			updateBindings();
		}
	}


	//////////////////////////////////////////////
	// Implementation of interface ItemListener //
	//////////////////////////////////////////////


	public void itemStateChanged(ItemEvent e)
	{
		switch (e.getStateChange())
		{
		case ItemEvent.SELECTED:
			viewBtn.setEnabled(true);
			unbindBtn.setEnabled(true);
			if (bindings[bindingsLst.getSelectedIndex()].type == NSbinding.CONTEXT)
			{
				downBtn.setEnabled(true);
				destroyBtn.setEnabled(true);
			}
			else
			{
				downBtn.setEnabled(false);
				destroyBtn.setEnabled(false);
			}
			break;
		case ItemEvent.DESELECTED:
			downBtn.setEnabled(false);
			viewBtn.setEnabled(false);
			unbindBtn.setEnabled(false);
			destroyBtn.setEnabled(false);
			break;
		}
	}


	///////////////////////////////////////////////////////////
	// inner WindowAdapter-derived class for the main window //
	///////////////////////////////////////////////////////////


	class OnWindowClosing extends WindowAdapter
	{
		public void windowClosing(WindowEvent e)
		{
			NSbrowser.this.dispose();
			if (standalone)
			{
				System.exit(0);
			}
		}
	}


	/////////////////////////////////////////////////
	// inner classes for the "bind" option's window //
	/////////////////////////////////////////////////


	class ContextChkMgr implements ItemListener
	{
		TextArea txt;

		public ContextChkMgr(TextArea the_txt)
		{
			txt = the_txt;
		}

		public void itemStateChanged(ItemEvent evt)
		{
			if (((Checkbox)evt.getSource()).getState())
			{
				txt.setEnabled(false);
			}
			else
			{
				txt.setEnabled(true);
			}
		}
	}

	class BindDlgMgr extends WindowAdapter
	{
		Dialog my_dlg;

		public BindDlgMgr(Dialog the_dlg)
		{
			my_dlg = the_dlg;
		}

		public void windowClosing(WindowEvent event)
		{
			my_dlg.dispose();
		}
	}

	class CancelBtnMgr implements ActionListener
	{
		Dialog my_dlg;

		public CancelBtnMgr(Dialog the_dlg)
		{
			my_dlg = the_dlg;
		}

		public void actionPerformed(ActionEvent event)
		{
			my_dlg.dispose();
		}
	}

	class BindBtnMgr implements ActionListener
	{
		Dialog bindDlg;
		TextField idFld, kindFld;
		Checkbox contextChk;
		TextArea refTxt;

		public BindBtnMgr(Dialog bind,
						  TextField id,
						  TextField kind,
						  Checkbox context,
						  TextArea ref)
		{
			bindDlg = bind;
			idFld = id;
			kindFld = kind;
			contextChk = context;
			refTxt = ref;
		}

		public void actionPerformed(ActionEvent event)
		{
			if (contextChk.getState())
			{
				try
				{
					NSbrowser.this.my_ns.makePath(
						NSbrowser.this.path + idFld.getText() + "!" + kindFld.getText());
				}
				catch (NameServiceException e)
				{
					new AlertWindow(NSbrowser.this, "Cannot create the new context", e.toString(), "OK");
				}
			}
			else
			{
				try
				{
					NSbrowser.this.my_ns.bind(
						NSbrowser.this.path + idFld.getText() + "!" + kindFld.getText(),
						NSbrowser.this.my_orb.string_to_object(refTxt.getText())
					);
				}
				catch (SystemException e)
				{
					new AlertWindow(NSbrowser.this, "Invalid Object Reference", e.toString(), "OK");
				}
				catch (NameServiceException e)
				{
					new AlertWindow(NSbrowser.this, "Name Service Exception", e.toString(), "OK");
				}
			}
			updateBindings();
			bindDlg.dispose();
		}
	}
}
