/*
* MobiliTools: an implementation of the Object Management Group's
* Mobile Agent Facility specification.
* Copyright (C) 2003 France Telecom R&D
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
* MobiliTools $Name:  $
*
* Contact: mobilitools-smi@lists.debian-sf.objectweb.org
*
* Authors: Bruno Dillenseger
*/


package org.objectweb.mobilitools.util.corba;


import org.omg.CORBA.*;
import org.omg.CosNaming.*;
import org.omg.CosNaming.NamingContextPackage.*;
import java.util.*;


/**
 * MobiliTools $Name:  $, $Id: NameService.java,v 1.1.1.1 2003/03/28 14:48:06 dillense Exp $
 * <P>
 * Front-end utility for the CORBA naming service.
 */
public class NameService
{
	// reference to the default name service
	NamingContext nameServiceRef;
	static final int LIST_MAX_ONCE = 100;


	/**
		Creates and initializes a new name service front-end.
		@param orb reference to the current ORB object (see return from ORB.init() call).
		@exception NameServiceException thrown if the initial reference to a naming service
			cannot be obtained.
	*/
	public NameService(ORB orb) throws NameServiceException
	{
		org.omg.CORBA.Object objRef = null;

		try
		{
			objRef = orb.resolve_initial_references("NameService");
		}
		catch (org.omg.CORBA.ORBPackage.InvalidName e)
		{
			throw new NameServiceException(e);
		}
		if (objRef != null)
		{
			try
			{
				nameServiceRef = NamingContextHelper.narrow(objRef);
			}
			catch (org.omg.CORBA.BAD_PARAM e)
			{
				throw new NameServiceException(e);
			}
		}
		else
		{
			throw new NameServiceException();
		}
	}


	/**
		Creates and initializes a new name service front-end.
		@param obj object reference narrowing to a naming context.
		@exception NameServiceException thrown if the object reference to a naming service
			cannot be narrowed to a naming context.
	*/
	public NameService(org.omg.CORBA.Object obj) throws NameServiceException
	{
		if (obj != null)
		{
			try
			{
				nameServiceRef = NamingContextHelper.narrow(obj);
			}
			catch (org.omg.CORBA.BAD_PARAM e)
			{
				throw new NameServiceException(e);
			}
		}
		else
		{
			throw new NameServiceException();
		}
	}


	/**
		Resolves a String-fied name to the bound object reference.
		@param name string translation ("context!kind/.../object!kind") of the object's name.
		@return the object reference bound to the named object.
		@exception NameServiceException same as resolve(org.omg.CosNaming.NameComponent[])
		@see #resolve(org.omg.CosNaming.NameComponent[])
		@see #stringToName(java.lang.String)
	*/
	public org.omg.CORBA.Object resolve(String name) throws NameServiceException
	{
		return resolve(stringToName(name));
	}


	/**
		Resolves a name to the bound object reference.
		@param name name of the object in the naming service.
		@return the object reference bound to the named object.
		@exception NameServiceException thrown in several failure situations:
			<UL>
				<LI>CORBA system exception,
				<LI>the name is invalid,
				<LI>the name was not found,
				<LI>naming service internal error.
			</UL>
			See the exception attribute of the NameServiceException, and refer to the CORBA naming service
			definitions to get the actual diagnostic.
	*/
	public org.omg.CORBA.Object resolve(NameComponent[] name) throws NameServiceException
	{
		org.omg.CORBA.Object obj = null;

		if (name.length == 0)
		{
			return nameServiceRef;
		}

		try
		{
			obj = nameServiceRef.resolve(name);
		}
		catch (SystemException e)
		{
			throw new NameServiceException(e);
		}
		catch (InvalidName e)
		{
			throw new NameServiceException(e);
		}
		catch (NotFound e)
		{
			throw new NameServiceException(e);
		}
		catch (CannotProceed e)
		{
			throw new NameServiceException(e);
		}
		return obj;
	}


	/**
		Returns an array of objects of contexts bound in the given context name.
		@param contextName the string-fied name of the context to list ("contex!kind/..." format).
		@return an array of NSbinding objects listing the objects and contexts found.
		@exception NameServiceException same as list(org.omg.CORBA.NameComponent[]).
		@see #list(NameComponent[])
	*/
	public NSbinding[] list(String contextName) throws NameServiceException
	{
		return list(stringToName(contextName));
	}


	/**
		Returns an array of objects of contexts bound in the given context name.
		@param contextName name of the context to list.
		@return an array of NSbinding objects listing the objects and contexts found.
		@exception NameServiceException thrown if a communication exception occurs
			while getting the content list of the specified context, or if the specified
			context cannot be narrowed to a naming context. In the former case, the actual
			exception object is attached to the NameServiceException. Since some bindings may
			disappear while resolving the object references, the resulting exceptions are
			ignored.
	*/
	public NSbinding[] list(NameComponent[] contextName) throws NameServiceException
	{
		NamingContext context = null;
		BindingListHolder listHolder = new BindingListHolder();
		BindingIteratorHolder iteratorHolder = new BindingIteratorHolder();
		NSbinding element;
		int size;

		try
		{
			context = NamingContextHelper.narrow(resolve(contextName));
		}
		catch (org.omg.CORBA.BAD_PARAM e)
		{
			throw new NameServiceException(e);
		}
		try
		{
			context.list(LIST_MAX_ONCE, listHolder, iteratorHolder);
		}
		catch (SystemException e)
		{
			throw new NameServiceException(e);
		}

		Vector vect = new Vector(LIST_MAX_ONCE, LIST_MAX_ONCE);
		boolean again = true;
		while (again)
		{
			for (int i=0 ; i<listHolder.value.length ; ++i)
			{
				try
				{
					element = new NSbinding(
						listHolder.value[i].binding_name[listHolder.value[i].binding_name.length-1].id,
						listHolder.value[i].binding_name[listHolder.value[i].binding_name.length-1].kind,
						(listHolder.value[i].binding_type.value() == BindingType._nobject
						? NSbinding.OBJECT : NSbinding.CONTEXT),
						context.resolve(listHolder.value[i].binding_name));
					vect.addElement(element);
				}
				catch  (SystemException e)
				{
					throw new NameServiceException(e);
				}
				catch (InvalidName e)
				{
				}
				catch (NotFound e)
				{
				}
				catch (CannotProceed e)
				{
					throw new NameServiceException(e);
				}
			}
			if (iteratorHolder.value != null)
			{
				try
				{
					again = iteratorHolder.value.next_n(LIST_MAX_ONCE, listHolder);
				}
				catch (SystemException e)
				{
					throw new NameServiceException(e);
				}
			}
			else
			{
				again = false;
			}
		}
		NSbinding[] result = new NSbinding[vect.size()];
		Enumeration lister = vect.elements();
		for (int i=0 ; lister.hasMoreElements() ; ++i)
		{
			result[i] = (NSbinding)lister.nextElement();
		}
		return result;
	}


	public void makePath(String name) throws NameServiceException
	{
		makePath(stringToName(name));
	}


	public void makePath(NameComponent[] name) throws NameServiceException
	{
		NamingContext nc = nameServiceRef;
		NameComponent[] step = new NameComponent[1];

		for (int i=0 ; i<name.length ; ++i)
		{
			step[0] = name[i];
			try
			{
				nc = nc.bind_new_context(step);
			}
			catch (AlreadyBound e)
			{
				try
				{
					nc = NamingContextHelper.narrow(nc.resolve(step));
				}
				catch (SystemException e2)
				{
					throw new NameServiceException(e);
				}
				catch (InvalidName e2)
				{
					throw new NameServiceException(e);
				}
				catch (NotFound e2)
				{
					throw new NameServiceException(e);
				}
				catch (CannotProceed e2)
				{
					throw new NameServiceException(e);
				}
			}
			catch  (SystemException e)
			{
				throw new NameServiceException(e);
			}
			catch (InvalidName e)
			{
				throw new NameServiceException(e);
			}
			catch (NotFound e)
			{
				throw new NameServiceException(e);
			}
			catch (CannotProceed e)
			{
				throw new NameServiceException(e);
			}
		}
	}


	public void bind(String name, org.omg.CORBA.Object obj) throws NameServiceException
	{
		bind(stringToName(name), obj, false);
	}


	public void bind(NameComponent[] name, org.omg.CORBA.Object obj) throws NameServiceException
	{
		bind(name, obj, false);
	}


	public void rebind(String name, org.omg.CORBA.Object obj) throws NameServiceException
	{
		bind(stringToName(name), obj, true);
	}


	public void rebind(NameComponent[] name, org.omg.CORBA.Object obj) throws NameServiceException
	{
		bind(name, obj, true);
	}


	public void unbind(String name) throws NameServiceException
	{
		unbind(stringToName(name));
	}


	public void unbind(NameComponent[] name) throws NameServiceException
	{
		try
		{
			nameServiceRef.unbind(name);
		}
		catch (SystemException e)
		{
			throw new NameServiceException(e);
		}
		catch (InvalidName e)
		{
			throw new NameServiceException(e);
		}
		catch (NotFound e)
		{
			throw new NameServiceException(e);
		}
		catch (CannotProceed e)
		{
			throw new NameServiceException(e);
		}
	}


	public void destroy(String name) throws NameServiceException
	{
		destroy(stringToName(name));
	}


	public void destroy(NameComponent[] name) throws NameServiceException
	{
		try
		{
			NamingContext nc = NamingContextHelper.narrow(nameServiceRef.resolve(name));
			unbind(name);
			nc.destroy();
		}
		catch (SystemException e)
		{
			throw new NameServiceException(e);
		}
		catch (InvalidName e)
		{
			throw new NameServiceException(e);
		}
		catch (NotFound e)
		{
			throw new NameServiceException(e);
		}
		catch (CannotProceed e)
		{
			throw new NameServiceException(e);
		}
		catch (NotEmpty e)
		{
			throw new NameServiceException(e);
		}
		catch (NullPointerException e)
		{
			throw new NameServiceException(e);
		}
	}


	public void deepUnbind(String name) throws NameServiceException
	{
		deepUnbind(stringToName(name));
	}


	public void deepUnbind(NameComponent[] name) throws NameServiceException
	{
		if (isContext(name))
		{
			NSbinding[] bindings = list(name);
			NameComponent[] dependency = new NameComponent[1 + name.length];
			int i;

			for (i=0 ; i<name.length ; ++i)
			{
				dependency[i] = name[i];
			}
			for (i=0 ; i<bindings.length ; ++i)
			{
				dependency[name.length] = new NameComponent(bindings[i].name, bindings[i].kind);
				deepUnbind(dependency);
			}
			destroy(name);
		}
		else
		{
			unbind(name);
		}
	}


	public void bind(NameComponent[] name, org.omg.CORBA.Object obj, boolean force)
		throws NameServiceException
	{
		if (name.length > 0)
		{
			NameComponent[] path = new NameComponent[name.length - 1];
			for (int i=0 ; i<path.length ; ++i)
			{
				path[i] = name[i];
			}
			makePath(path);
			NameComponent[] singleName = new NameComponent[1];
			singleName[0] = name[name.length-1];
			NamingContext context = NamingContextHelper.narrow(resolve(path));
			try
			{
				if (force)
				{
					if (isContext(obj))
					{
						context.rebind_context(singleName, NamingContextHelper.narrow(obj));
					}
					else
					{
						context.rebind(singleName, obj);
					}
				}
				else
				{
					if (isContext(obj))
					{
						context.bind_context(singleName, NamingContextHelper.narrow(obj));
					}
					else
					{
						context.bind(singleName, obj);
					}
				}
			}
			catch (SystemException e)
			{
				throw new NameServiceException(e);
			}
			catch (InvalidName e)
			{
				throw new NameServiceException(e);
			}
			catch (NotFound e)
			{
				throw new NameServiceException(e);
			}
			catch (CannotProceed e)
			{
				throw new NameServiceException(e);
			}
			catch (AlreadyBound e)
			{
				throw new NameServiceException(e);
			}
		}
		else
		{
			throw new NameServiceException("The provided name is empty.");
		}
	}


	static public boolean isContext(org.omg.CORBA.Object obj)
	{
		if (obj instanceof NamingContext)
		{
			return true;
		}
		try
		{
			NamingContextHelper.narrow(obj);
		}
		catch (org.omg.CORBA.BAD_PARAM e)
		{
		 	return false;
		}
		return true;
	}


	public boolean isContext(String name) throws NameServiceException
	{
		return NameService.isContext(resolve(name));
	}


	public boolean isContext(NameComponent[] name) throws NameServiceException
	{
		return NameService.isContext(resolve(name));
	}


	static public String nameToString(NameComponent[] name)
	{
		String strName = "";

		for (int i=0 ; i<name.length ; ++i)
		{
			strName = strName + name[i].id;
			if (name[i].kind.length() != 0)
			{
				strName = strName + "!" + name[i].kind;
			}
			strName = strName + "/";
		}
		return strName;
	}


	static public NameComponent[] stringToName(String strName)
	{
		StringTokenizer parser = new StringTokenizer(strName, "/", false);
		NameComponent[] name = new NameComponent[parser.countTokens()];
		StringTokenizer subParser;
		String token;

		for (int i=0 ; i<name.length ; ++i)
		{
			subParser = new StringTokenizer(parser.nextToken(), "!", true);
			name[i] = new NameComponent();
			if (subParser.hasMoreTokens())
			{
				token = subParser.nextToken();
				if (token.equals("!"))
				{
					name[i].id = "";
				}
				else
				{
					name[i].id = token;
				}
				if (subParser.hasMoreTokens())
				{
					token = subParser.nextToken();
					if (token.equals("!"))
					{
						if (subParser.hasMoreTokens())
						{
							name[i].kind = subParser.nextToken();
						}
						else
						{
							name[i].kind = "";
						}
					}
					else
					{
						name[i].kind = token;
					}
				}
				else
				{
					name[i].kind = "";
				}
			}
			else
			{
				name[i].id = "";
				name[i].kind = "";
			}
		}
		return name;
	}
}
