/*
* MobiliTools: an implementation of the Object Management Group's
* Mobile Agent Facility specification.
* Copyright (C) 2003 France Telecom R&D
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
* MobiliTools $Name:  $
*
* Contact: mobilitools-smi@lists.debian-sf.objectweb.org
*
* Authors: Bruno Dillenseger
*/


package org.objectweb.mobilitools.util.thread;


/**
 * MobiliTools $Name:  $, $Id: ShutdownLock.java,v 1.1.1.1 2003/03/28 14:48:06 dillense Exp $
 * <P>
 * Sort of one-shot RW-lock. Many "reader" threads can enter (<code>in()</code>)
 * and then quit (<code>out()</code>) the lock as long as <code>shutdown()</code>
 * has not been invoked. Once <code>shutdown()</code> is called by a thread,
 * other threads trying to enter the lock will get a ShutdownException.
 * <shutdown()> invocation returns as soon as every thread has quitted the lock.
*/
public class ShutdownLock
{
	volatile int in = 0;
	volatile Thread thread = null;
	volatile boolean shutdown = false;
	boolean trace = false;
	String name = null;


	/**
		Creates a new "shutdown lock" with no name and no traces.
	*/
	public ShutdownLock()
	{
	}


	/**
		Creates a new "shutdown lock" with the provided name, with activated traces.
		@param lockName name that will be displayed to identify the lock in traces.
	*/
	public ShutdownLock(String lockName)
	{
		name = lockName;
		trace = true;
	}


	/**
		Enters the lock. Don't forget to call <code>out()</code>
		once and only once per each <code>in()</code> call.
		@throws ShutdownException if the lock has been shutdown.
	*/
	synchronized public void in()
		throws ShutdownException
	{
		if (shutdown && Thread.currentThread() != thread)
		{
			throw new ShutdownException();
		}
		++in;
		if (trace)
		{
			System.out.println(Thread.currentThread() + " entered ShutdownLock " + name + " (=" + in + ").");
		}
	}


	/**
		Quits the lock. Must be called once and only once after each
		<code>in()</code> invocation.
	*/
	synchronized public void out()
	{
		if (--in == 0)
		{
			if (trace)
			{
				System.out.println("ShutdownLock " + name + " is released.");
			}
			if (shutdown && Thread.currentThread() != thread)
			{
				notify();
			}
		}
		else if (trace)
		{
			System.out.println("ShutdownLock " + name + " = " + in + ".");
		}
	}


	/**
		Shuts the lock down. Any other thread trying to enter
		the lock will get a ShutdownException exception.
		Returns as soon as every other thread has quitted the lock.
		@throws ShutdownException if the lock has already been shut down.
	*/
	synchronized public void shutdown()
		throws ShutdownException
	{
		if (shutdown)
		{
			throw new ShutdownException();
		}
		shutdown = true;
		thread = Thread.currentThread();
		if (in > 0)
		{
			if (trace)
			{
				System.out.println("ShutdownLock " + name + " shutdown is pending for " + thread);
			}
			try
			{
				wait();
			}
			catch (InterruptedException e)
			{
				e.printStackTrace();
			}
		}
		if (trace)
		{
			System.out.println("ShutdownLock " + name + " is shutdown.");
		}
	}


	/**
		Exception thrown to mean that the invoked lock has been shutdown.
	*/
	public class ShutdownException extends Exception
	{
		public ShutdownException()
		{
		}
		public ShutdownException(String msg)
		{
			super(msg);
		}
	}
}
